#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4:et:sw=4:

# ----------------------------------------------------------------------
# Copyleft (K), Jose M. Rodriguez-Rosa (a.k.a. Boriel)
#
# This program is Free Software and is released under the terms of
#                    the GNU General License
# ----------------------------------------------------------------------

from api.constants import TYPE
from api.constants import CLASS
from api.config import OPTIONS
from api.decorator import classproperty
from .symbol_ import Symbol


class SymbolTYPE(Symbol):
    """ A Type definition. Defines a type,
    both user defined or basic ones.
    """

    def __init__(self, name, lineno, *children):
        # All children (if any) must be SymbolTYPE
        assert all(isinstance(x, SymbolTYPE) for x in children)
        super(SymbolTYPE, self).__init__(*children)
        self.name = name  # typename
        self.lineno = lineno  # The line the type was defined. Line 0 = basic type
        self.final = self  # self.final always return the original aliased type (if this type is an alias)
        self.caseins = OPTIONS.case_insensitive.value  # Whether this ID is case insensitive or not
        self.class_ = CLASS.type_
        self.accessed = False  # Whether this type has been used or not

    def __repr__(self):
        return "%s(%s)" % (self.token, str(self))

    def __str__(self):
        return self.name

    @property
    def size(self):
        return sum(x.size for x in self.children)

    @property
    def is_basic(self):
        """ Whether this is a basic (canonical) type or not.
        """
        if len(self.children) == 1:
            return self.children[0].is_basic

        return False

    @property
    def is_signed(self):
        if self is not self.final:
            return self.final.is_signed

        if len(self.children) != 1:
            return False

        return self.children[0].is_signed

    @property
    def is_dynamic(self):
        """ True if this type uses dynamic (Heap) memory.
        e.g. strings or dynamic arrays
        """
        if self is not self.final:
            return self.final.is_dynamic

        return any([x.is_dynamic for x in self.children])

    @property
    def is_alias(self):
        """ Whether this is an alias of another type or not.
        """
        return False

    def __eq__(self, other):
        assert isinstance(other, SymbolTYPE)

        if self is not self.final:
            return self.final == other

        other = other.final  # remove alias

        if other.is_basic:
            return other == self

        if len(self.children) != len(other.children):
            if len(self.children) == 1 and not other.children:
                return self.children[0] == other
            if len(other.children) == 1 and not self.children:
                return other.children[0] == self
            return False

        return all(i == j for i, j in zip(self.children, other.children))

    def __ne__(self, other):
        assert isinstance(other, SymbolTYPE)
        return not (self == other)

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        if self is not self.final:
            return bool(self.final)

        return any(x for x in self.children)


class SymbolBASICTYPE(SymbolTYPE):
    """ Defines a basic type (Ubyte, Byte, etc..)
    Basic (default) types are defined upon start and are case insensitive.
    If name is None or '', default typename from TYPES.to_string will be used.
    """

    def __init__(self, type_, name=None):
        """ type_ = Internal representation (e.g. TYPE.ubyte)
        """
        assert TYPE.is_valid(type_)
        if not name:
            name = TYPE.to_string(type_)

        super(SymbolBASICTYPE, self).__init__(name, 0)
        self.type_ = type_

    @property
    def size(self):
        return TYPE.size(self.type_)

    @property
    def is_basic(self):
        """ Whether this is a basic (canonical) type or not.
        """
        return True

    @property
    def is_signed(self):
        return TYPE.is_signed(self.type_)

    def to_signed(self):
        """ Returns another instance with the signed equivalent
        of this type.
        """
        return SymbolBASICTYPE(TYPE.to_signed(self.type_))

    @property
    def is_dynamic(self):
        return self.type_ == TYPE.string

    def __hash__(self):
        return self.type_

    def __eq__(self, other):
        if self is not self.final:
            return self.final == other

        other = other.final  # remove alias
        if other.is_basic:  # for both basic types, just compare
            return self.type_ == other.type_

        assert other.children  # must be not empty
        if len(other.children) > 1:  # Different size
            return False

        return self == other.children[0]

    def __bool__(self):
        return self.type_ != TYPE.unknown


class SymbolTYPEALIAS(SymbolTYPE):
    """ Defines a type which is alias of another
    """

    def __init__(self, name, lineno, alias):
        assert isinstance(alias, SymbolTYPE)
        super(SymbolTYPEALIAS, self).__init__(name, lineno, alias)
        self.final = alias.final

    @property
    def is_alias(self):
        """ Whether this is an alias of another type or not.
        """
        return True

    @property
    def size(self):
        return self.final.size

    @property
    def is_basic(self):
        return self.final.is_basic

    @property
    def alias(self):
        return self.children[0]

    @property
    def to_signed(self):
        assert self.is_basic
        return self.final.to_signed()


class SymbolTYPEREF(SymbolTYPEALIAS):
    """ Creates a Type reference or usage.
    Eg. DIM a As Integer
    In this case, the Integer type is accessed.
    It's an alias type, containing just the
    original Type definition (SymbolTYPE), the
    the lineno it is currently being accessed,
    and if it was implicitly inferred or explicitly declared.
    """

    def __init__(self, type_, lineno, implicit=False):
        assert isinstance(type_, SymbolTYPE)
        super(SymbolTYPEREF, self).__init__(type_.name, lineno, type_)
        self.implicit = implicit

    def to_signed(self):
        assert self.is_basic
        return self.final.to_signed()

    @property
    def type_(self):
        assert self.is_basic
        return self.final.type_


class Type(object):
    """ Class for enumerating Basic Types.
    e.g. Type.string.
    """
    unknown = auto = SymbolBASICTYPE(TYPE.unknown)
    ubyte = SymbolBASICTYPE(TYPE.ubyte)
    byte_ = SymbolBASICTYPE(TYPE.byte_)
    uinteger = SymbolBASICTYPE(TYPE.uinteger)
    long_ = SymbolBASICTYPE(TYPE.long_)
    ulong = SymbolBASICTYPE(TYPE.ulong)
    integer = SymbolBASICTYPE(TYPE.integer)
    fixed = SymbolBASICTYPE(TYPE.fixed)
    float_ = SymbolBASICTYPE(TYPE.float_)
    string = SymbolBASICTYPE(TYPE.string)

    types = [unknown,
             ubyte, byte_,
             uinteger, integer,
             ulong, long_,
             fixed, float_,
             string]

    _by_name = {x.name: x for x in types}

    @staticmethod
    def size(t):
        assert isinstance(t, SymbolTYPE)
        return t.size

    @staticmethod
    def to_string(t):
        assert isinstance(t, SymbolTYPE)
        return t.name

    @classmethod
    def by_name(cls, typename):
        """ Converts a given typename to Type
        """
        return cls._by_name.get(typename, None)

    @classproperty
    def integrals(cls):
        return (cls.byte_, cls.ubyte, cls.integer, cls.uinteger,
                cls.long_, cls.ulong)

    @classproperty
    def signed(cls):
        return cls.byte_, cls.integer, cls.long_, cls.fixed, cls.float_

    @classproperty
    def unsigned(cls):
        return cls.ubyte, cls.uinteger, cls.ulong

    @classproperty
    def decimals(cls):
        return cls.fixed, cls.float_

    @classproperty
    def numbers(cls):
        return tuple(list(cls.integrals) + list(cls.decimals))

    @classmethod
    def is_numeric(cls, t):
        assert isinstance(t, SymbolTYPE)
        return t.final in cls.numbers

    @classmethod
    def is_signed(cls, t):
        assert isinstance(t, SymbolTYPE)
        return t.final in cls.signed

    @classmethod
    def is_unsigned(cls, t):
        assert isinstance(t, SymbolTYPE)
        return t.final in cls.unsigned

    @classmethod
    def is_integral(cls, t):
        assert isinstance(t, SymbolTYPE)
        return t.final in cls.integrals

    @classmethod
    def is_decimal(cls, t):
        assert isinstance(t, SymbolTYPE)
        return t.final in cls.decimals

    @classmethod
    def is_string(cls, t):
        assert isinstance(t, SymbolTYPE)
        return t.final == cls.string

    @classmethod
    def to_signed(cls, t):
        """ Return signed type or equivalent
        """
        assert isinstance(t, SymbolTYPE)
        t = t.final
        assert t.is_basic
        if cls.is_unsigned(t):
            return {cls.ubyte: cls.byte_,
                    cls.uinteger: cls.integer,
                    cls.ulong: cls.long_}[t]
        if cls.is_signed(t) or cls.is_decimal(t):
            return t
        return cls.unknown



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from unittest import TestCase
from symbols import NOP


__author__ = 'boriel'


class TestSymbolBLOCK(TestCase):
    def setUp(self):
        self.nop = NOP()

    def test__len_0(self):
        self.assertEqual(len(self.nop), 0, "NOP must have 0 length")

    def test__assert_false(self):
        self.assertFalse(self.nop)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:et:

# ----------------------------------------------------------------------
# Copyleft (K), Jose M. Rodriguez-Rosa (a.k.a. Boriel)
#
# This program is Free Software and is released under the terms of
#                    the GNU General License
# ----------------------------------------------------------------------

from . import config
from . import global_
from .constants import CLASS
from .constants import SCOPE
import api.errmsg
from .errmsg import error


__all__ = ['check_type',
           'check_is_declared_explicit',
           'check_type_is_explicit',
           'check_call_arguments',
           'check_pending_calls',
           'check_pending_labels',
           'is_number',
           'is_const',
           'is_static',
           'is_string',
           'is_numeric',
           'is_dynamic',
           'is_null',
           'is_unsigned',
           'common_type'
           ]


# ----------------------------------------------------------------------
# Several check functions.
# These functions trigger syntax errors if checking goal fails.
# ----------------------------------------------------------------------


def check_type(lineno, type_list, arg):
    """ Check arg's type is one in type_list, otherwise,
    raises an error.
    """
    if not isinstance(type_list, list):
        type_list = [type_list]

    if arg.type_ in type_list:
        return True

    if len(type_list) == 1:
        error(lineno, "Wrong expression type '%s'. Expected '%s'" %
              (arg.type_, type_list[0]))
    else:
        error(lineno, "Wrong expression type '%s'. Expected one of '%s'"
              % (arg.type_, tuple(type_list)))

    return False


def check_is_declared_explicit(lineno, id_, classname='variable'):
    """ Check if the current ID is already declared.
    If not, triggers a "undeclared identifier" error,
    if the --explicit command line flag is enabled (or #pragma
    option strict is in use).

    If not in strict mode, passes it silently.
    """
    if not config.OPTIONS.explicit.value:
        return True

    entry = global_.SYMBOL_TABLE.check_is_declared(id_, lineno, classname)
    return entry is not None  # True if declared


def check_type_is_explicit(lineno: int, id_: str, type_):
    from symbols.type_ import SymbolTYPE
    assert isinstance(type_, SymbolTYPE)
    if type_.implicit:
        if config.OPTIONS.strict.value:
            api.errmsg.syntax_error_undeclared_type(lineno, id_)


def check_call_arguments(lineno, id_, args):
    """ Check arguments against function signature.

        Checks every argument in a function call against a function.
        Returns True on success.
    """
    if not global_.SYMBOL_TABLE.check_is_declared(id_, lineno, 'function'):
        return False

    if not global_.SYMBOL_TABLE.check_class(id_, CLASS.function, lineno):
        return False

    entry = global_.SYMBOL_TABLE.get_entry(id_)

    if len(args) != len(entry.params):
        c = 's' if len(entry.params) != 1 else ''
        error(lineno, "Function '%s' takes %i parameter%s, not %i" %
              (id_, len(entry.params), c, len(args)))
        return False

    for arg, param in zip(args, entry.params):
        if arg.class_ in (CLASS.var, CLASS.array) and param.class_ != arg.class_:
            error(lineno, "Invalid argument '{}'".format(arg.value))
            return None

        if not arg.typecast(param.type_):
            return False

        if param.byref:
            from symbols.var import SymbolVAR
            if not isinstance(arg.value, SymbolVAR):
                error(lineno, "Expected a variable name, not an expression (parameter By Reference)")
                return False

            if arg.class_ not in (CLASS.var, CLASS.array):
                error(lineno, "Expected a variable or array name (parameter By Reference)")
                return False

            arg.byref = True

        if arg.value is not None:
            arg.value.add_required_symbol(param)

    if entry.forwarded:  # The function / sub was DECLARED but not implemented
        error(lineno, "%s '%s' declared but not implemented" % (CLASS.to_string(entry.class_), entry.name))
        return False

    return True


def check_pending_calls():
    """ Calls the above function for each pending call of the current scope
    level
    """
    result = True

    # Check for functions defined after calls (parameters, etc)
    for id_, params, lineno in global_.FUNCTION_CALLS:
        result = result and check_call_arguments(lineno, id_, params)

    return result


def check_pending_labels(ast):
    """ Iteratively traverses the node looking for ID with no class set,
    marks them as labels, and check they've been declared.

    This way we avoid stack overflow for high line-numbered listings.
    """
    result = True
    visited = set()
    pending = [ast]

    while pending:
        node = pending.pop()

        if node is None or node in visited:  # Avoid recursive infinite-loop
            continue

        visited.add(node)
        for x in node.children:
            pending.append(x)

        if node.token != 'VAR' or (node.token == 'VAR' and node.class_ is not CLASS.unknown):
            continue

        tmp = global_.SYMBOL_TABLE.get_entry(node.name)
        if tmp is None or tmp.class_ is CLASS.unknown:
            error(node.lineno, 'Undeclared identifier "%s"'
                  % node.name)
        else:
            assert tmp.class_ == CLASS.label
            node.to_label(node)

        result = result and tmp is not None

    return result


def check_and_make_label(lbl, lineno):
    """ Checks if the given label (or line number) is valid and, if so,
    returns a label object.
    :param lbl: Line number of label (string)
    :param lineno: Line number in the basic source code for error reporting
    :return: Label object or None if error.
    """
    if isinstance(lbl, float):
        if lbl == int(lbl):
            id_ = str(int(lbl))
        else:
            error(lineno, 'Line numbers must be integers.')
            return None
    else:
        id_ = lbl

    return global_.SYMBOL_TABLE.access_label(id_, lineno)


# ----------------------------------------------------------------------
# Function for checking some arguments
# ----------------------------------------------------------------------
def is_null(*symbols):
    """ True if no nodes or all the given nodes are either
    None, NOP or empty blocks. For blocks this applies recursively
    """
    from symbols.symbol_ import Symbol

    for sym in symbols:
        if sym is None:
            continue
        if not isinstance(sym, Symbol):
            return False
        if sym.token == 'NOP':
            continue
        if sym.token == 'BLOCK':
            if not is_null(*sym.children):
                return False
            continue
        return False
    return True


def is_SYMBOL(token, *symbols):
    """ Returns True if ALL of the given argument are AST nodes
    of the given token (e.g. 'BINARY')
    """
    from symbols.symbol_ import Symbol
    assert all(isinstance(x, Symbol) for x in symbols)
    for sym in symbols:
        if sym.token != token:
            return False

    return True


def is_LABEL(*p):
    return is_SYMBOL('LABEL', *p)


def is_string(*p):
    return is_SYMBOL('STRING', *p)


def is_const(*p):
    """ A constant in the program, like CONST a = 5
    """
    return is_SYMBOL('VAR', *p) and all(x.class_ == CLASS.const for x in p)


def is_CONST(*p):
    """ Not to be confused with the above.
    Check it's a CONSTant expression
    """
    return is_SYMBOL('CONST', *p)


def is_static(*p):
    """ A static value (does not change at runtime)
    which is known at compile time
    """
    return all(is_CONST(x) or
               is_number(x) or
               is_const(x)
               for x in p)


def is_number(*p):
    """ Returns True if ALL of the arguments are AST nodes
    containing NUMBER or numeric CONSTANTS
    """
    try:
        for i in p:
            if i.token != 'NUMBER' and (i.token != 'ID' or i.class_ != CLASS.const):
                return False

        return True
    except:
        pass

    return False


def is_var(*p):
    """ Returns True if ALL of the arguments are AST nodes
    containing ID
    """
    return is_SYMBOL('VAR', *p)


def is_integer(*p):
    from symbols.type_ import Type

    try:
        for i in p:
            if not i.is_basic or not Type.is_integral(i.type_):
                return False

        return True
    except:
        pass

    return False


def is_unsigned(*p):
    """ Returns false unless all types in p are unsigned
    """
    from symbols.type_ import Type

    try:
        for i in p:
            if not i.type_.is_basic or not Type.is_unsigned(i.type_):
                return False

        return True
    except:
        pass

    return False


def is_signed(*p):
    """ Returns false unless all types in p are signed
    """
    from symbols.type_ import Type

    try:
        for i in p:
            if not i.type_.is_basic or not Type.is_signed(i.type_):
                return False

        return True
    except:
        pass

    return False


def is_numeric(*p):
    """ Returns false unless all elements in p are of numerical type
    """
    from symbols.type_ import Type

    try:
        for i in p:
            if not i.type_.is_basic or not Type.is_numeric(i.type_):
                return False

        return True
    except:
        pass

    return False


def is_type(type_, *p):
    """ True if all args have the same type
    """
    try:
        for i in p:
            if i.type_ != type_:
                return False

        return True
    except:
        pass

    return False


def is_dynamic(*p):  # TODO: Explain this better
    """ True if all args are dynamic (e.g. Strings, dynamic arrays, etc)
    The use a ptr (ref) and it might change during runtime.
    """
    from symbols.type_ import Type

    try:
        for i in p:
            if i.scope == SCOPE.global_ and i.is_basic and \
                    i.type_ != Type.string:
                return False

        return True
    except:
        pass

    return False


def is_callable(*p):
    """ True if all the args are functions and / or subroutines
    """
    import symbols
    return all(isinstance(x, symbols.FUNCTION) for x in p)


def is_block_accessed(block):
    """ Returns True if a block is "accessed". A block of code is accessed if
    it has a LABEL and it is used in a GOTO, GO SUB or @address access
    :param block: A block of code (AST node)
    :return: True / False depending if it has labels accessed or not
    """
    if is_LABEL(block) and block.accessed:
        return True

    for child in block.children:
        if not is_callable(child) and is_block_accessed(child):
            return True

    return False


def is_temporary_value(node) -> bool:
    """ Returns if the AST node value is a variable or a temporary copy in the heap.
    """
    return node.token not in ('STRING', 'VAR') and node.t[0] not in ('_', '#')


def common_type(a, b):
    """ Returns a type which is common for both a and b types.
    Returns None if no common types allowed.
    """
    from symbols.type_ import SymbolBASICTYPE as BASICTYPE
    from symbols.type_ import Type as TYPE
    from symbols.type_ import SymbolTYPE

    if a is None or b is None:
        return None

    if not isinstance(a, SymbolTYPE):
        a = a.type_

    if not isinstance(b, SymbolTYPE):
        b = b.type_

    if a == b:  # Both types are the same?
        return a  # Returns it

    if a == TYPE.unknown and b == TYPE.unknown:
        return BASICTYPE(global_.DEFAULT_TYPE)

    if a == TYPE.unknown:
        return b

    if b == TYPE.unknown:
        return a

    # TODO: This will removed / expanded in the future
    assert a.is_basic
    assert b.is_basic

    types = (a, b)

    if TYPE.float_ in types:
        return TYPE.float_

    if TYPE.fixed in types:
        return TYPE.fixed

    if TYPE.string in types:  # TODO: Check this ??
        return TYPE.unknown

    result = a if a.size > b.size else b

    if not TYPE.is_unsigned(a) or not TYPE.is_unsigned(b):
        result = TYPE.to_signed(result)

    return result



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------
# Copyleft (K), Jose M. Rodriguez-Rosa (a.k.a. Boriel)
#
# This program is Free Software and is released under the terms of
#                    the GNU General License
#
# This is the Lexer for the ZXBppASM (ZXBASM Preprocessor)
# ----------------------------------------------------------------------

import os
import sys
from ply import lex
import api.utils
from libzxbpp.prepro.output import warning, error

EOL = '\n'

# Names for std input/output
STDOUT = '<stdout>'
STDIN = '<stdin>'
STDERR = '<stderr>'

states = (
    ('prepro', 'exclusive'),
    ('define', 'exclusive'),
    ('defargsopt', 'exclusive'),
    ('defargs', 'exclusive'),
    ('defexpr', 'exclusive'),
    ('pragma', 'exclusive'),
    ('singlecomment', 'exclusive'),
    ('asmcomment', 'exclusive')
)

_tokens = ('STRING', 'TOKEN', 'NEWLINE', '_ENDFILE_', 'FILENAME', 'ID',
           'INTEGER', 'EQ', 'PUSH', 'POP', 'LP', 'LLP', 'RRP', 'RP', 'COMMA',
           'CONTINUE', 'NUMBER', 'SEPARATOR'
           )

reserved_directives = {
    'include': 'INCLUDE',
    'once': 'ONCE',
    'define': 'DEFINE',
    'undef': 'UNDEF',
    'ifdef': 'IFDEF',
    'ifndef': 'IFNDEF',
    'else': 'ELSE',
    'endif': 'ENDIF',
    'init': 'INIT',
    'line': 'LINE',
    'require': 'REQUIRE',
    'pragma': 'PRAGMA',
}

# List of token names.
tokens = _tokens + tuple(reserved_directives.values())

__COMMENT_LEVEL = 0


class Lexer(object):
    """ Own class lexer to allow multiple instances.
    This lexer is just a wrapper of the current FILESTACK[-1] lexer
    """

    # -------------- TOKEN ACTIONS --------------
    def t_INITIAL_asmcomment_CONTINUE(self, t):
        r'[\\_]\r?\n'
        t.lexer.lineno += 1
        return t

    def t_INITIAL_COMMENT(self, t):
        r';'
        t.lexer.push_state('asmcomment')
        t.type = 'TOKEN'
        t.value = ';'
        return t

    def t_asmcomment_TOKEN(self, t):
        r'.+'
        return t

    def t_asmcomment_NEWLINE(self, t):
        r'\r?\n'
        # New line => remove whatever state in top of the stack and replace it with INITIAL
        t.lexer.lineno += 1
        t.lexer.pop_state()
        return t

    def t_INITIAL_NEWLINE(self, t):
        r'\r?\n'
        # New line => remove whatever state in top of the stack and replace it with INITIAL
        t.lexer.lineno += 1
        return t

    def t_INITIAL_ID(self, t):
        r'[_A-Za-z][_A-Za-z0-9]*'
        return t

    def t_INITIAL_CHAR(self, t):
        r"'([^'\n]|'')'"
        t.type = 'TOKEN'
        return t

    def t_INITIAL_TOKEN(self, t):
        r"[][%'`,.:$()*/<>~&|+^-]"
        return t

    def t_prepro_define_defargs_defargsopt_defexpr_pragma_NEWLINE(self, t):
        r'\r?\n'
        t.lexer.lineno += 1
        t.lexer.pop_state()
        return t

    def t_prepro_define_defargs_defargsopt_defexpr_pragma_COMMENT(self, t):
        r";"
        t.lexer.begin('singlecomment')

    def t_singlecomment_NEWLINE(self, t):
        r'\r?\n'
        t.lexer.pop_state()  # Back to initial
        t.lexer.lineno += 1
        return t

    def t_prepro_define_pragma_defargs_defargsopt_CONTINUE(self, t):
        r'[_\\]\r?\n'
        t.lexer.lineno += 1
        t.value = t.value[1:]
        t.type = 'NEWLINE'
        return t

    # Any other character is ignored until EOL
    def t_singlecomment_comment_Skip(self, t):
        r'.+'
        pass

    # Allows line breaking
    def t_defexpr_CONTINUE(self, t):
        r'[\\_]\r?\n'
        t.lexer.lineno += 1
        return t

    def t_prepro_pragma_defargs_define_skip(self, t):
        r'[ \t]+'
        pass  # Ignore whitespaces and tabs

    def t_prepro_ID(self, t):
        r'[_a-zA-Z][_a-zA-Z0-9]*'  # preprocessor directives
        t.type = reserved_directives.get(t.value.lower(), 'ID')
        if t.type == 'DEFINE':
            t.lexer.begin('define')
        elif t.type == 'PRAGMA':
            t.lexer.begin('pragma')
        return t

    def t_pragma_LP(self, t):
        r'\('
        return t

    def t_pragma_RP(self, t):
        r'\)'
        return t

    def t_defexpr_LLP(self, t):
        r'\('
        return t

    def t_defexpr_RRP(self, t):
        r'\)'
        return t

    def t_pragma_ID(self, t):
        r'[_a-zA-Z][_a-zA-Z0-9]*'  # pragma directives
        if t.value.upper() in ('PUSH', 'POP'):
            t.type = t.value.upper()
        return t

    def t_defargsopt_LP(self, t):
        r'\('
        t.lexer.begin('defargs')
        return t

    # Any other char than '(' means no arglist
    def t_defargsopt_TOKEN(self, t):
        r'[ \t)@,%&~|{}:.+*/-]|=>|<=|<>|=|<|>|<<|>>'
        t.lexer.begin('defexpr')
        return t

    def t_defargsopt_STRING(self, t):
        r'"([^"\n]|"")*"'  # a doubled quoted string
        t.lexer.begin('defexpr')
        return t

    def t_defargs_LP(self, t):
        r'\('
        return t

    def t_defargs_RP(self, t):
        r'\)'
        t.lexer.begin('defexpr')
        return t

    def t_defargsopt_defexpr_ID(self, t):
        r'[_a-zA-Z][_a-zA-Z0-9]*'  # preprocessor directives
        t.lexer.begin('defexpr')
        return t

    def t_defargs_defargsopt_CONTINUE(self, t):
        r'[\\_]\r?\n'
        t.lexer.lineno += 1
        pass

    def t_defargsopt_SEPARATOR(self, t):
        r'[ \t]+'
        t.lexer.begin('defexpr')

    def t_defargs_ID(self, t):
        r'[_a-zA-Z][_a-zA-Z0-9]*'  # preprocessor directives
        return t

    def t_define_ID(self, t):
        r'[_a-zA-Z][_a-zA-Z0-9]*'  # preprocessor directives
        t.lexer.begin('defargsopt')
        return t

    def t_prepro_pragma_INTEGER(self, t):
        r'[0-9]+'  # an integer number
        return t

    def t_prepro_pragma_STRING(self, t):
        r'"([^"]|"")*"'  # a doubled quoted string
        t.value = t.value[1:-1]  # Remove quotes
        return t

    def t_defexpr_INITIAL_STRING(self, t):
        r'"([^"]|"")*"'  # a doubled quoted string
        return t

    def t_pragma_EQ(self, t):
        r'='
        return t

    def t_defexpr_defargs_prepro_COMMA(self, t):
        r','
        return t

    def t_INIIAL_sharp(self, t):
        r'\#'  # Only matches if at beginning of line and "#"
        if t.value == '#' and self.find_column(t) == 1:
            t.lexer.push_state('prepro')  # Start preprocessor
        else:
            t.type = 'TOKEN'
            return t

    def t_defexpr_TOKEN(self, t):
        r'=>|<=|>=|<>|[@:.<>^=+*/&|%-]'
        return t

    def t_defexpr_INITIAL_SEPARATOR(self, t):
        r'[ \t]+'
        return t

    def t_defexpr_INITIAL_HEXA(self, t):
        r'([0-9][0-9a-fA-F]*[hH])|(\$[0-9a-fA-F]+)'
        # Hexadecimal numbers
        t.type = 'NUMBER'
        return t

    def t_defexpr_INITIAL_BIN(self, t):
        r'(%[01]+)|([01]+[bB])'  # A Binary integer
        # Note 00B is a 0 binary, but
        # 00Bh is a 12 in hex. So this pattern must come
        # after HEXA
        t.type = 'NUMBER'
        return t

    def t_defexpr_INITIAL_NUMBER(self, t):
        # This pattern must come AFTER t_HEXA and t_BIN
        r'(([0-9]+(\.[0-9]+)?)|(\.[0-9]+))([eE][-+]?[0-9]+)?'
        return t

    def t_prepro_FILENAME(self, t):
        r'<[^>]*>'
        t.value = t.value[1:-1]  # Remove quotes
        return t

    def t_defargs_defargsopt_prepro_define_defexpr_pragma_singlecomment_INITIAL_asmcomment_ANY(self, t):
        r'.'
        self.error("illegal preprocessor character '%s'" % t.value[0])

    def t_defargs_defargsopt_prepro_define_defexpr_pragma_singlecomment_INITIAL_asmcomment_error(self, t):
        """ error handling rule. This should never happens!
        """
        pass  # The lexer will raise an exception here. This is intended

    def put_current_line(self, prefix=''):
        """ Returns line and file for include / end of include sequences.
        """
        return '%s#line %i "%s"\n' % (prefix, self.lex.lineno, os.path.basename(self.filestack[-1][0]))

    def include(self, filename):
        """ Changes FILENAME and line count
        """
        if filename != STDIN and filename in [x[0] for x in self.filestack]:  # Already included?
            self.warning(' Recursive inclusion')

        self.filestack.append([filename, 1, self.lex, self.input_data])
        self.lex = lex.lex(object=self)
        result = self.put_current_line()  # First #line start with \n (EOL)

        try:
            if filename == STDIN:
                self.input_data = sys.stdin.read()
            else:
                self.input_data = api.utils.read_txt_file(filename)

            if len(self.input_data) and self.input_data[-1] != EOL:
                self.input_data += EOL
        except IOError:
            self.input_data = EOL

        self.lex.input(self.input_data)
        return result

    def include_end(self):
        """ Performs and end of include.
        """
        self.lex = self.filestack[-1][2]
        self.input_data = self.filestack[-1][3]
        self.filestack.pop()

        if not self.filestack:  # End of input?
            return

        self.filestack[-1][1] += 1  # Increment line counter of previous file

        result = lex.LexToken()  # Creates the token
        result.value = self.put_current_line()
        result.type = '_ENDFILE_'
        result.lineno = self.lex.lineno
        result.lexpos = self.lex.lexpos

        return result

    def input(self, str_, filename=''):
        """ Defines input string, removing current lexer.
        """
        self.filestack.append([filename, 1, self.lex, self.input_data])

        self.input_data = str_
        self.lex = lex.lex(object=self)
        self.lex.input(self.input_data)

    def token(self):
        """ Returns a token from the current input. If tok is None
        from the current input, it means we are at end of current input
        (e.g. at end of include file). If so, closes the current input
        and discards it; then pops the previous input and lexer from
        the input stack, and gets another token.

        If new token is again None, repeat the process described above
        until the token is either not None, or self.lex is None, wich
        means we must effectively return None, because parsing has
        ended.
        """
        tok = None
        if self.next_token is not None:
            tok = lex.LexToken()
            tok.value = ''
            tok.lineno = self.lex.lineno
            tok.lexpos = self.lex.lexpos
            tok.type = self.next_token
            self.next_token = None

        while self.lex is not None and tok is None:
            tok = self.lex.token()
            if tok is not None:
                break

            tok = self.include_end()

        return tok

    def find_column(self, token):
        """ Compute column:
                - token is a token instance
        """
        i = token.lexpos
        while i > 0:
            if self.input_data[i - 1] == '\n':
                break
            i -= 1

        column = token.lexpos - i + 1

        return column

    def error(self, msg):
        """ Prints an error msg and continues execution.
        """
        error(self.lex.lineno, msg)

    def warning(self, msg):
        """ Emits a warning and continue execution.
        """
        warning(self.lex.lineno, msg)

    def __init__(self):
        """ Creates a new GLOBAL lexer instance
        """
        self.lex = None
        self.filestack = []  # Current filename, and line number being parsed
        self.input_data = ''
        self.tokens = tokens
        self.states = states
        self.next_token = None  # if set to something, this will be returned once


# --------------------- PREPROCESSOR FUNCTIONS -------------------

# Needed for states
tmp = lex.lex(object=Lexer())

# ------------------ Test if called from cmd line ---------------
if __name__ == '__main__':  # For testing purposes
    lexer = Lexer()
    lexer.include(sys.argv[1])

    while 1:
        tok = lexer.token()
        if not tok:
            break
        print(tok)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import sys
from api import config
from api import global_


class TestConfig(unittest.TestCase):
    """ Tests api.config initialization
    """
    def setUp(self):
        config.OPTIONS.reset()

    def test_init(self):
        config.init()
        self.assertEqual(config.OPTIONS.Debug.value, 0)
        self.assertEqual(config.OPTIONS.stdin.value, sys.stdin)
        self.assertEqual(config.OPTIONS.stdout.value, sys.stdout)
        self.assertEqual(config.OPTIONS.stderr.value, sys.stderr)
        self.assertEqual(config.OPTIONS.optimization.value, global_.DEFAULT_OPTIMIZATION_LEVEL)
        self.assertEqual(config.OPTIONS.case_insensitive.value, False)
        self.assertEqual(config.OPTIONS.array_base.value, 0)
        self.assertEqual(config.OPTIONS.byref.value, False)
        self.assertEqual(config.OPTIONS.max_syntax_errors.value, global_.DEFAULT_MAX_SYNTAX_ERRORS)
        self.assertEqual(config.OPTIONS.string_base.value, 0)
        self.assertEqual(config.OPTIONS.memory_map.value, None)
        self.assertEqual(config.OPTIONS.bracket.value, False)
        self.assertEqual(config.OPTIONS.use_loader.value, False)
        self.assertEqual(config.OPTIONS.autorun.value, False)
        self.assertEqual(config.OPTIONS.output_file_type.value, 'bin')
        self.assertEqual(config.OPTIONS.include_path.value, '')
        self.assertEqual(config.OPTIONS.memoryCheck.value, False)
        self.assertEqual(config.OPTIONS.strictBool.value, False)
        self.assertEqual(config.OPTIONS.arrayCheck.value, False)
        self.assertEqual(config.OPTIONS.enableBreak.value, False)
        self.assertEqual(config.OPTIONS.emitBackend.value, False)
        self.assertEqual(config.OPTIONS.arch.value, 'zx48k')
        # private options that cannot be accessed with #pragma
        self.assertEqual(config.OPTIONS.option('__DEFINES').value, {})
        self.assertEqual(config.OPTIONS.explicit.value, False)
        self.assertEqual(config.OPTIONS.Sinclair.value, False)
        self.assertEqual(config.OPTIONS.strict.value, False)

    def test_initted_values(self):
        config.init()
        self.assertEqual(sorted(config.OPTIONS.options.keys()), ['Debug',
                                                                 'Sinclair',
                                                                 'StdErrFileName',
                                                                 '__DEFINES',
                                                                 'arch',
                                                                 'arrayCheck',
                                                                 'array_base',
                                                                 'autorun',
                                                                 'bracket',
                                                                 'byref',
                                                                 'case_insensitive',
                                                                 'emitBackend',
                                                                 'enableBreak',
                                                                 'explicit',
                                                                 'include_path',
                                                                 'inputFileName',
                                                                 'max_syntax_errors',
                                                                 'memoryCheck',
                                                                 'memory_map',
                                                                 'optimization',
                                                                 'outputFileName',
                                                                 'output_file_type',
                                                                 'stderr',
                                                                 'stdin',
                                                                 'stdout',
                                                                 'strict',
                                                                 'strictBool',
                                                                 'string_base',
                                                                 'use_loader',
                                                                 'zxnext'])



# -*- coding: utf-8 -*-

import re
import itertools

RE_SVAR = re.compile(r'(\$(?:\$|[0-9]+))')
RE_PARSE = re.compile(r'(\s+|"(?:[^"]|"")*")')


class BasicLinePattern(object):
    """ Defines a pattern for a line, like 'push $1' being
    $1 a pattern variable
    """
    @staticmethod
    def sanitize(pattern):
        """ Returns a sanitized pattern version of a string to be later
        compiled into a reg exp
        """
        meta = r'.^$*+?{}[]\|()'
        return ''.join(r'\%s' % x if x in meta else x for x in pattern)

    def __init__(self, line):
        self.line = ''.join(x.strip() or ' ' for x in RE_PARSE.split(line) if x).strip()
        self.vars = []
        self.re_pattern = ''
        self.output = []

        for token in RE_PARSE.split(self.line):
            if token == ' ':
                self.re_pattern += r'\s+'
                self.output.append(' ')
                continue
            subtokens = [x for x in RE_SVAR.split(token) if x]

            for tok in subtokens:
                if tok == '$$':
                    self.re_pattern += r'\$'
                    self.output.append('$')
                elif RE_SVAR.match(tok):
                    self.output.append(tok)
                    mvar = '_%s' % tok[1:]
                    if mvar not in self.vars:
                        self.vars.append(mvar)
                        self.re_pattern += '(?P<%s>.*)' % mvar
                    else:
                        self.re_pattern += r'\%i' % (self.vars.index(mvar) + 1)
                else:
                    self.output.append(tok)
                    self.re_pattern += BasicLinePattern.sanitize(tok)

        self.re = re.compile(self.re_pattern)
        self.vars = set(x.replace('_', '$') for x in self.vars)


class LinePattern(BasicLinePattern):
    """ Defines a pattern to match against a source assembler.
    Given an assembler instruction with substitution variables
    ($1, $2, ...) creates an instance that matches against a list real
    assembler instructions. e.g.
      push $1
    matches against
      push af
    and bounds $1 to 'af'
    Note that $$ matches against the $ sign
    """
    def match(self, line, vars_=None):
        match = self.re.match(line)
        if match is None:
            return None

        vars__ = {'_%s' % k[1:]: v for k, v in (vars_ or {}).items()}
        for k, v in vars__.items():
            if match.groupdict().get(k, v) != v:
                return None

        result = dict(vars_ or {})
        result.update({'$%s' % k[1:]: v for k, v in match.groupdict().items()})
        return result

    def __repr__(self):
        return repr(self.re)


class BlockPattern(object):
    """ Given a list asm instructions, tries to match them
    """
    def __init__(self, lines):
        lines = [x.strip() for x in lines]
        self.patterns = [LinePattern(x) for x in lines if x]
        self.lines = [pattern.line for pattern in self.patterns]
        self.vars = set(itertools.chain(*[p.vars for p in self.patterns]))

    def __len__(self):
        return len(self.lines)

    def match(self, instructions, start=0):
        """ Given a list of instructions and a starting point,
        returns whether this pattern matches or not from such point
        onwards.
        E.g. given the pattern:
          push $1
          pop $1
        and the list
          ld a, 5
          push af
          pop af
        this pattern will match at position 1
        """
        lines = instructions[start:]
        if len(self) > len(lines):
            return None

        univars = {}
        for patt, line in zip(self.patterns, lines):
            univars = patt.match(line, vars_=univars)
            if univars is None:
                return None

        return univars

    def __repr__(self):
        return str([repr(x) for x in self.patterns])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4:et:sw=4:

# ----------------------------------------------------------------------
# Copyleft (K), Jose M. Rodriguez-Rosa (a.k.a. Boriel)
#
# This program is Free Software and is released under the terms of
#                    the GNU General License
# ----------------------------------------------------------------------

import os

from .decorator import classproperty


# -------------------------------------------------
# Global constants
# -------------------------------------------------

# Path to main ZX Basic compiler executable
ZXBASIC_ROOT = os.path.abspath(os.path.join(
    os.path.abspath(os.path.dirname(os.path.abspath(__file__))), os.path.pardir)
)

# ----------------------------------------------------------------------
# Class enums
# ----------------------------------------------------------------------


class CLASS(object):
    """ Enums class constants
    """
    unknown = 'unknown'  # 0
    var = 'var'  # 1  # scalar variable
    array = 'array'  # 2  # array variable
    function = 'function'  # 3  # function
    label = 'label'  # 4 Labels
    const = 'const'  # 5  # constant literal value e.g. 5 or "AB"
    sub = 'sub'  # 6  # subroutine
    type_ = 'type'  # 7  # type

    _CLASS_NAMES = {
        unknown: '(unknown)',
        var: 'var',
        array: 'array',
        function: 'function',
        label: 'label',
        const: 'const',
        sub: 'sub',
        type_: 'type'
    }

    @classproperty
    def classes(cls):
        return (cls.unknown, cls.var, cls.array, cls.function, cls.sub,
                cls.const, cls.label)

    @classmethod
    def is_valid(cls, class_):
        """ Whether the given class is
        valid or not.
        """
        return class_ in cls.classes

    @classmethod
    def to_string(cls, class_):
        assert cls.is_valid(class_)
        return cls._CLASS_NAMES[class_]


class ARRAY(object):
    """ Enums array constants
    """
    bound_size = 2  # This might change depending on arch, program, etc..
    bound_count = 2  # Size of bounds counter
    array_type_size = 1  # Size of array type


class TYPE(object):
    """ Enums type constants
    """
    auto = unknown = None
    byte_ = 1
    ubyte = 2
    integer = 3
    uinteger = 4
    long_ = 5
    ulong = 6
    fixed = 7
    float_ = 8
    string = 9

    TYPE_SIZES = {
        byte_: 1, ubyte: 1,
        integer: 2, uinteger: 2,
        long_: 4, ulong: 4,
        fixed: 4, float_: 5,
        string: 2, unknown: 0
    }

    TYPE_NAMES = {
        byte_: 'byte', ubyte: 'ubyte',
        integer: 'integer', uinteger: 'uinteger',
        long_: 'long', ulong: 'ulong',
        fixed: 'fixed', float_: 'float',
        string: 'string', unknown: 'none'
    }

    @classproperty
    def types(cls):
        return tuple(cls.TYPE_SIZES.keys())

    @classmethod
    def size(cls, type_):
        return cls.TYPE_SIZES.get(type_, None)

    @classproperty
    def integral(cls):
        return (cls.byte_, cls.ubyte, cls.integer, cls.uinteger,
                cls.long_, cls.ulong)

    @classproperty
    def signed(cls):
        return (cls.byte_, cls.integer, cls.long_, cls.fixed, cls.float_)

    @classproperty
    def unsigned(cls):
        return (cls.ubyte, cls.uinteger, cls.ulong)

    @classproperty
    def decimals(cls):
        return (cls.fixed, cls.float_)

    @classproperty
    def numbers(cls):
        return tuple(list(cls.integral) + list(cls.decimals))

    @classmethod
    def is_valid(cls, type_):
        """ Whether the given type is
        valid or not.
        """
        return type_ in cls.types

    @classmethod
    def is_signed(cls, type_):
        return type_ in cls.signed

    @classmethod
    def is_unsigned(cls, type_):
        return type_ in cls.unsigned

    @classmethod
    def to_signed(cls, type_):
        """ Return signed type or equivalent
        """
        if type_ in cls.unsigned:
            return {TYPE.ubyte: TYPE.byte_,
                    TYPE.uinteger: TYPE.integer,
                    TYPE.ulong: TYPE.long_}[type_]
        if type_ in cls.decimals or type_ in cls.signed:
            return type_
        return cls.unknown

    @classmethod
    def to_string(cls, type_):
        """ Return ID representation (string) of a type
        """
        return cls.TYPE_NAMES[type_]

    @classmethod
    def to_type(cls, typename):
        """ Converts a type ID to name. On error returns None
        """
        NAME_TYPES = {cls.TYPE_NAMES[x]: x for x in cls.TYPE_NAMES}
        return NAME_TYPES.get(typename, None)


class SCOPE(object):
    """ Enum scopes
    """
    unknown = None
    global_ = 'global'
    local = 'local'
    parameter = 'parameter'

    _names = {
        unknown: 'unknown',
        global_: 'global',
        local: 'local',
        parameter: 'parameter'
    }

    @classmethod
    def is_valid(cls, scope):
        return cls._names.get(scope, None) is not None

    @classmethod
    def to_string(cls, scope):
        assert cls.is_valid(scope)
        return cls._names[scope]


class KIND(object):
    """ Enum kind
    """
    unknown = None
    var = 'var'
    function = 'function'
    sub = 'sub'
    type_ = 'type'

    _NAMES = {
        unknown: '(unknown)',
        var: 'variable',
        function: 'function',
        sub: 'subroutine',
        type_: 'type'
    }

    @classmethod
    def is_valid(cls, kind):
        return cls._NAMES.get(kind, None) is not None

    @classmethod
    def to_string(cls, kind):
        assert cls.is_valid(kind)
        return cls._NAMES.get(kind)


class CONVENTION(object):
    unknown = None
    fastcall = '__fastcall__'
    stdcall = '__stdcall__'

    _NAMES = {
        unknown: '(unknown)',
        fastcall: '__fastcall__',
        stdcall: '__stdcall__'
    }

    @classmethod
    def is_valid(cls, convention):
        return cls._NAMES.get(convention, None) is not None

    @classmethod
    def to_string(cls, convention):
        assert cls.is_valid(convention)
        return cls._NAMES[convention]


# ----------------------------------------------------------------------
# Identifier Class (variable, function, label, array)
# ----------------------------------------------------------------------
ID_CLASSES = CLASS.classes

# ----------------------------------------------------------------------
# Deprecated suffixes for variable names, such as "a$"
# ----------------------------------------------------------------------
DEPRECATED_SUFFIXES = ('$', '%', '&')

# ----------------------------------------------------------------------
# Identifier type
# i8 = Integer, 8 bits
# u8 = Unsigned, 8 bits and so on
# ----------------------------------------------------------------------
ID_TYPES = TYPE.types

# Maps deprecated suffixes to types
SUFFIX_TYPE = {'$': TYPE.string, '%': TYPE.integer, '&': TYPE.long_}



# -*- coding: utf-8 -*-

import re

from . import patterns
from . import common

# All 'single' registers (even f FLAG one). SP is not decomposable so it's 'single' already
ALL_REGS = {'a', 'b', 'c', 'd', 'e', 'f', 'h', 'l',
            'ixh', 'ixl', 'iyh', 'iyl', 'r', 'i', 'sp'}

# The set of all registers as they can appear in any instruction as operands
REGS_OPER_SET = {'a', 'b', 'c', 'd', 'e', 'h', 'l',
                 'bc', 'de', 'hl', 'sp', 'ix', 'iy', 'ixh', 'ixl', 'iyh', 'iyl',
                 'af', "af'", 'i', 'r'}

# Instructions that marks the end of a basic block (any branching instruction)
BLOCK_ENDERS = {'jr', 'jp', 'call', 'ret', 'reti', 'retn', 'djnz', 'rst'}

UNKNOWN_PREFIX = '*UNKNOWN_'
END_PROGRAM_LABEL = '__END_PROGRAM'  # Label for end program
RE_UNK_PREFIX = re.compile('^' + re.escape(UNKNOWN_PREFIX) + r'\d+$')
HL_SEP = '|'  # Hi/Low separator


def new_tmp_val():
    """ Generates an 8-bit unknown value
    """
    common.RAND_COUNT += 1
    return '{0}{1}'.format(UNKNOWN_PREFIX, common.RAND_COUNT)


def new_tmp_val16():
    """ Generates an unknown 16-bit tmp value concatenating two 8-it unknown ones
    """
    return '{}{}{}'.format(new_tmp_val(), HL_SEP, new_tmp_val())


def is_unknown(x):
    if x is None:
        return True

    if isinstance(x, int):
        return False

    assert isinstance(x, str)
    xx = x.split(HL_SEP)
    if len(xx) > 2:
        return False

    return any(RE_UNK_PREFIX.match(_) for _ in xx)


def is_unknown8(x):
    if x is None:
        return True

    if not is_unknown(x):
        return False

    return len(x.split(HL_SEP)) == 1


def is_unknown16(x):
    if x is None:
        return True

    if not is_unknown(x):
        return False

    return len(x.split(HL_SEP)) == 2


def get_L_from_unknown_value(tmp_val):
    """ Given a 16bit *UNKNOWN value, returns it's lower part, which is the same 2nd part,
    after splitting by HL_SEP. If the parameter is None, a new tmp_value will be generated.
    If the value is a composed one (xxxH | yyyL) returns yyyL.
    """
    assert is_unknown(tmp_val), "Malformed unknown value '{}'".format(tmp_val)

    if tmp_val is None:
        tmp_val = new_tmp_val16()

    return tmp_val.split(HL_SEP)[-1]


def get_H_from_unknown_value(tmp_val):
    """ Given a 16bit *UNKNOWN value, returns it's higher part, which is the same 1st part,
    after splitting by HL_SEP. If the parameter is None, a new tmp_value will be generated.
    If the value is a composed one (xxxH | yyyL) returns yyyH.
    """
    assert is_unknown(tmp_val), "Malformed unknown value '{}'".format(tmp_val)

    if tmp_val is None:
        tmp_val = new_tmp_val16()

    return tmp_val.split(HL_SEP)[0]


def is_mem_access(arg):
    """ Returns if a given string is a memory access, that is
    if it matches the form (...)
    """
    arg = arg.strip()
    return (arg[0], arg[-1]) == ('(', ')')


# TODO: to be rewritten
def is_number(x):
    """ Returns whether X """
    if x is None or x == '':
        return False

    if isinstance(x, (int, float)):
        return True

    if isinstance(x, str):
        x = x.strip()

    if isinstance(x, str) and is_mem_access(x):
        return False

    try:
        tmp = eval(x, {}, {})
        if isinstance(tmp, (int, float)):
            return True
    except NameError:
        pass
    except SyntaxError:
        pass
    except ValueError:
        pass

    return patterns.RE_NUMBER.match(str(x)) is not None


def valnum(x):
    """ If x is a numeric value (int, float) or it's a string
    representation of a number (hexa, binary), returns it numeric value.
    Otherwise returns None
    """
    if not is_number(x):
        return None

    x = str(x)

    if x[0] == '%':
        return int(x[1:], 2)

    if x[-1] in ('b', 'B'):
        return int(x[:-1], 2)

    if x[0] == '$':
        return int(x[1:], 16)

    if x[-1] in ('h', 'H'):
        return int(x[:-1], 16)

    return int(eval(x, {}, {}))


def simplify_arg(arg):
    """ Given an asm operand (str), if it can be evaluated to a single 16 bit integer number it will be done so.
    Memory addresses will preserve their parenthesis. If the string can not be simplified, it will be
    returned as is.

    eg.:
        0       -> 0
        (0)     -> (0)
        0 + 3   -> 3
        (3 + 1) -> (4)
        (a - 1) -> (a - 1)
        b - 5   -> b - 5

    This is very simple "parsing" (for speed) and it won't understand (5) + (6) and will be returned as (11)
    """
    result = None
    arg = arg.strip()
    try:
        tmp = eval(arg, {}, {})
        if isinstance(tmp, (int, float)):
            result = str(tmp)
    except NameError:
        pass
    except SyntaxError:
        pass
    except ValueError:
        pass

    if result is None:
        return arg

    if not is_mem_access(arg):
        return result

    return '({})'.format(result)


def simplify_asm_args(asm):
    """ Given an asm instruction try to simplify its args.
    """
    chunks = [x for x in asm.split(' ', 1)]
    if len(chunks) != 2:
        return asm

    args = [simplify_arg(x) for x in chunks[1].split(',', 1)]
    return '{} {}'.format(chunks[0], ', '.join(args))


def is_register(x):
    """ True if x is a register.
    """
    if not isinstance(x, str):
        return False

    return x.lower() in REGS_OPER_SET


def is_8bit_normal_register(x):
    """ Returns whether the given string x is a "normal" 8 bit register. Those are 8 bit registers
    which belongs to the normal (documented) Z80 instruction set as operands (so a', f', ixh, etc
    are excluded).
    """
    return x.lower() in {'a', 'b', 'c', 'd', 'e', 'i', 'h', 'l'}


def is_8bit_idx_register(x):
    """ Returns whether the given string x one of the undocumented IX, IY 8 bit registers.
    """
    return x.lower() in {'ixh', 'ixl', 'iyh', 'iyl'}


def is_8bit_oper_register(x):
    """ Returns whether the given string x is an 8 bit register that can be used as an
    instruction operand. This included those of the undocumented Z80 instruction set as
    operands (ixh, ixl, etc) but not h', f'.
    """
    return x.lower() in {'a', 'b', 'c', 'd', 'e', 'i', 'h', 'l', 'ixh', 'ixl', 'iyh', 'iyl'}


def is_16bit_normal_register(x):
    """ Returns whether the given string x is a "normal" 16 bit register. Those are 16 bit registers
    which belongs to the normal (documented) Z80 instruction set as operands which can be operated
    directly (i.e. load a value directly), and not for indexation (IX + n, for example).
    So af, ix, iy, sp, bc', hl', de' are excluded.
    """
    return x.lower() in {'bc', 'de', 'hl'}


def is_16bit_idx_register(x):
    """ Returns whether the given string x is a indexable (i.e. IX + n) 16 bit register.
    """
    return x.lower() in {'ix', 'iy'}


def is_16bit_composed_register(x):
    """ A 16bit register that can be decomposed into a high H16 and low L16 part
    """
    return x.lower() in {'af', "af'", 'bc', 'de', 'hl', 'ix', 'iy'}


def is_16bit_oper_register(x):
    """ Returns whether the given string x is a 16 bit register. These are any 16 bit register
    which belongs to the normal (documented) Z80 instruction set as operands.
    """
    return x.lower() in {'af', "af'", 'bc', 'de', 'hl', 'ix', 'iy', 'sp'}


def LO16(x):
    """ Given a 16-bit register (lowercase string), returns the low 8 bit register of it.
    The string *must* be a 16 bit lowercase register. SP register is not "decomposable" as
    two 8-bit registers and this is considered an error.
    """
    x = x.lower()
    assert is_16bit_oper_register(x), "'%s' is not a 16bit register" % x
    assert x != 'sp', "'sp' register cannot be split into two 8 bit registers"

    if is_16bit_idx_register(x):
        return x + 'l'

    return x[1] + ("'" if "'" in x else '')


def HI16(x):
    """ Given a 16-bit register (lowercase string), returns the high 8 bit register of it.
    The string *must* be a 16 bit lowercase register. SP register is not "decomposable" as
    two 8-bit registers and this is considered an error.
    """
    x = x.lower()
    assert is_16bit_oper_register(x), "'%s' is not a 16bit register" % x
    assert x != 'sp', "'sp' register cannot be split into two 8 bit registers"

    if is_16bit_idx_register(x):
        return x + 'h'

    return x[0] + ("'" if "'" in x else '')


def single_registers(op):
    """ Given an iterable (set, list) of registers like ['a', 'bc', "af'", 'h', 'hl'] returns
    a list of single registers: ['a', "a'", "f'", 'b', 'c', 'h', 'l'].
    Non register parameters (like numbers) will be ignored.

    Notes:
        - SP register will be returned as is since it's not decomposable in two 8 bit registers.
        - IX and IY will be returned as {'ixh', 'ixl'} and {'iyh', 'iyl'} respectively
    """
    result = set()
    if not isinstance(op, (list, set)):
        op = [op]

    for x in op:
        if is_8bit_oper_register(x) or x.lower() in ('f', 'sp'):
            result.add(x)
        elif not is_16bit_oper_register(x):
            continue
        else:  # Must be a 16bit reg or we have an internal error!
            result = result.union([LO16(x), HI16(x)])

    return sorted(result)


def idx_args(x):
    """ Given an argument x (string), returns None if it's not an index operation "ix/iy + n"
    Otherwise return a tuple (reg, oper, offset). It's case insensitive and the register is always returned
    in lowercase.

    Notice the parenthesis must NOT be included. So '(ix + 5)' won't match, whilst 'ix + 5' will.

    For example:
     - 'ix + 3' => ('ix', '+', '3')
     - 'IY - Something + 4' => ('iy', '-', 'Something + 4')
    """
    match = patterns.RE_IDX.match(x)
    if match is None:
        return None

    reg, sign, args = match.groups()
    return reg.lower(), sign, args


def LO16_val(x):
    """ Given an x value, it must be None, unknown, or an integer.
    Then returns it lower part. If it's none, a new tmp will be returned.
    """
    if x is None:
        return new_tmp_val()

    if valnum(x) is not None:
        return str(int(x) & 0xFF)

    if not is_unknown(x):
        return new_tmp_val()

    return x.split(HL_SEP)[-1]


def HI16_val(x):
    """ Given an x value, it must be None, unknown, or an integer.
    Then returns it upper part. If it's None, a new tmp will be returned.
    It it's an unknown8, return 0, because it's considered an 8 bit value.
    """
    if x is None:
        return new_tmp_val()

    if valnum(x) is not None:
        return str((int(x) >> 8) & 0xFF)

    if not is_unknown(x):
        return new_tmp_val()

    return "0{}{}".format(HL_SEP, x).split(HL_SEP)[-2]


def dict_intersection(dict_a, dict_b):
    """ Given 2 dictionaries a, b, returns a new one which contains the common key/pair values.
    e.g. for {'a': 1, 'b': 'x'}, {'a': 'q', 'b': 'x', 'c': 2} returns {'b': 'x'}

    :param dict_a: a python dictionary (or compatible class instance)
    :param dict_b: a python dictionary (or compatible class instance)

    :return a python dictionary with the key/val intersection
    """
    return {k: v for k, v in dict_a.items() if k in dict_b and dict_b[k] == v}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import symbols
from symbols.type_ import Type


class TestSymbolSTRING(TestCase):
    def test__init__(self):
        self.assertRaises(AssertionError, symbols.STRING, 0, 1)
        _zxbasic = 'zxbasic'
        _ZXBASIC = 'ZXBASIC'
        s = symbols.STRING(_zxbasic, 1)
        t = symbols.STRING(_ZXBASIC, 2)
        self.assertEqual(s, s)
        self.assertNotEqual(s, t)
        self.assertEqual(s, _zxbasic)
        self.assertEqual(_ZXBASIC, t)
        self.assertGreater(s, t)
        self.assertLess(t, s)
        self.assertGreaterEqual(s, t)
        self.assertLessEqual(t, s)
        self.assertEqual(s.type_, Type.string)
        self.assertEqual(str(s), _zxbasic)
        self.assertEqual('"{}"'.format(_zxbasic), s.__repr__())
        self.assertEqual(s.t, _zxbasic)
        s.t = _ZXBASIC
        self.assertEqual(s.t, _ZXBASIC)
        self.assertRaises(AssertionError, symbols.STRING.t.fset, s, 0)
        self.assertEqual(s.value, _zxbasic)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# vim: ts=4:et:sw=4:

# ----------------------------------------------------------------------
# Copyleft (K), Jose M. Rodriguez-Rosa (a.k.a. Boriel)
#
# This program is Free Software and is released under the terms of
#                    the GNU General License
# ----------------------------------------------------------------------

from api import global_
import api.symboltable
from .symbol_ import Symbol
from .function import SymbolFUNCTION


class SymbolFUNCDECL(Symbol):
    """ Defines a Function declaration
    """
    def __init__(self, entry, lineno):
        super(SymbolFUNCDECL, self).__init__()
        self.entry = entry  # Symbol table entry
        self.lineno = lineno  # Line of this function declaration

    @property
    def entry(self):
        return self.children[0]

    @entry.setter
    def entry(self, value):
        assert isinstance(value, SymbolFUNCTION)
        self.children = [value]

    @property
    def name(self):
        return self.entry.name

    @property
    def locals_size(self):
        return self.entry.locals_size

    @locals_size.setter
    def locals_size(self, value):
        self.entry.locals_size = value

    @property
    def local_symbol_table(self):
        return self.entry.local_symbol_table

    @local_symbol_table.setter
    def local_symbol_table(self, value):
        assert isinstance(value, api.symboltable.SymbolTable.Scope)
        self.entry.local_symbol_table = value

    @property
    def type_(self):
        return self.entry.type_

    @type_.setter
    def type_(self, value):
        self.entry.type_ = value

    @property
    def size(self):
        return self.type_.size

    @property
    def mangled(self):
        return self.entry.mangled

    @classmethod
    def make_node(cls, func_name, lineno, type_=None):
        """ This will return a node with the symbol as a function.
        """
        entry = global_.SYMBOL_TABLE.declare_func(func_name, lineno, type_=type_)
        if entry is None:
            return None

        entry.declared = True
        return cls(entry, lineno)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# vim:ts=4:et:sw=4:

# --------------------------------------------------------------
# Copyleft (k) 2008, by Jose M. Rodriguez-Rosa
# (a.k.a. Boriel, http://www.boriel.com)
#
# This module contains local array (both parameters and
# comparation intermediate-code traductions
# --------------------------------------------------------------
from api import fp
from .__common import REQUIRES
from .__float import _fpush
from .__f16 import f16


def _paddr(offset):
    """ Generic element array-address stack-ptr loading.
    Emits output code for setting IX at the right location.
    bytes = Number of bytes to load:
        1 => 8 bit value
        2 => 16 bit value / string
        4 => 32 bit value / f16 value
        5 => 40 bit value
    """
    output = []

    indirect = offset[0] == '*'
    if indirect:
        offset = offset[1:]

    I = int(offset)
    if I >= 0:
        I += 4  # Return Address + "push IX"

    output.append('push ix')
    output.append('pop hl')
    output.append('ld de, %i' % I)
    output.append('add hl, de')

    if indirect:
        output.append('call __ARRAY_PTR')
    else:
        output.append('call __ARRAY')
    REQUIRES.add('array.asm')
    return output


def _paaddr(ins):
    """ Loads address of an array element into the stack
    """
    output = _paddr(ins.quad[2])
    output.append('push hl')

    return output


def _paload8(ins):
    """ Loads an 8 bit value from a memory address
    If 2nd arg. start with '*', it is always treated as
    an indirect value.
    """
    output = _paddr(ins.quad[2])
    output.append('ld a, (hl)')
    output.append('push af')

    return output


def _paload16(ins):
    """ Loads a 16 bit value from a memory address
    If 2nd arg. start with '*', it is always treated as
    an indirect value.
    """
    output = _paddr(ins.quad[2])

    output.append('ld e, (hl)')
    output.append('inc hl')
    output.append('ld d, (hl)')
    output.append('ex de, hl')
    output.append('push hl')

    return output


def _paload32(ins):
    """ Loads a 32 bit value from a memory address
    If 2nd arg. start with '*', it is always treated as
    an indirect value.
    """
    output = _paddr(ins.quad[2])

    output.append('call __ILOAD32')
    output.append('push de')
    output.append('push hl')

    REQUIRES.add('iload32.asm')

    return output


def _paloadf(ins):
    """ Loads a floating point value from a memory address.
    If 2nd arg. start with '*', it is always treated as
    an indirect value.
    """
    output = _paddr(ins.quad[2])
    output.append('call __ILOADF')
    output.extend(_fpush())

    REQUIRES.add('iloadf.asm')

    return output


def _paloadstr(ins):
    """ Loads a string value from a memory address.
    """
    output = _paddr(ins.quad[2])

    output.append('call __ILOADSTR')
    output.append('push hl')
    REQUIRES.add('loadstr.asm')

    return output


def _pastore8(ins):
    """ Stores 2º operand content into address of 1st operand.
    1st operand is an array element. Dimensions are pushed into the
    stack.
    Use '*' for indirect store on 1st operand (A pointer to an array)
    """
    output = _paddr(ins.quad[1])

    value = ins.quad[2]
    if value[0] == '*':
        value = value[1:]
        indirect = True
    else:
        indirect = False

    try:
        value = int(ins.quad[2]) & 0xFFFF
        if indirect:
            output.append('ld a, (%i)' % value)
            output.append('ld (hl), a')
        else:
            value &= 0xFF
            output.append('ld (hl), %i' % value)
    except ValueError:
        output.append('pop af')
        output.append('ld (hl), a')

    return output


def _pastore16(ins):
    """ Stores 2º operand content into address of 1st operand.
    store16 a, x =>  *(&a) = x
    Use '*' for indirect store on 1st operand.
    """
    output = _paddr(ins.quad[1])

    value = ins.quad[2]
    if value[0] == '*':
        value = value[1:]
        indirect = True
    else:
        indirect = False

    try:
        value = int(ins.quad[2]) & 0xFFFF
        output.append('ld de, %i' % value)
        if indirect:
            output.append('call __LOAD_DE_DE')
            REQUIRES.add('lddede.asm')

    except ValueError:
        output.append('pop de')

    output.append('ld (hl), e')
    output.append('inc hl')
    output.append('ld (hl), d')

    return output


def _pastore32(ins):
    """ Stores 2º operand content into address of 1st operand.
    store16 a, x =>  *(&a) = x
    """
    output = _paddr(ins.quad[1])

    value = ins.quad[2]
    if value[0] == '*':
        value = value[1:]
        indirect = True
    else:
        indirect = False

    try:
        value = int(ins.quad[2]) & 0xFFFFFFFF  # Immediate?
        if indirect:
            output.append('push hl')
            output.append('ld hl, %i' % (value & 0xFFFF))
            output.append('call __ILOAD32')
            output.append('ld b, h')
            output.append('ld c, l')  # BC = Lower 16 bits
            output.append('pop hl')
            REQUIRES.add('iload32.asm')
        else:
            output.append('ld de, %i' % (value >> 16))
            output.append('ld bc, %i' % (value & 0xFFFF))
    except ValueError:
        output.append('pop bc')
        output.append('pop de')

    output.append('call __STORE32')
    REQUIRES.add('store32.asm')

    return output


def _pastoref16(ins):
    """ Stores 2º operand content into address of 1st operand.
    storef16 a, x =>  *(&a) = x
    """
    output = _paddr(ins.quad[1])

    value = ins.quad[2]
    if value[0] == '*':
        value = value[1:]
        indirect = True
    else:
        indirect = False

    try:
        if indirect:
            value = int(ins.quad[2])
            output.append('push hl')
            output.append('ld hl, %i' % (value & 0xFFFF))
            output.append('call __ILOAD32')
            output.append('ld b, h')
            output.append('ld c, l')  # BC = Lower 16 bits
            output.append('pop hl')
            REQUIRES.add('iload32.asm')
        else:
            de, hl = f16(value)
            output.append('ld de, %i' % de)
            output.append('ld bc, %i' % hl)
    except ValueError:
        output.append('pop bc')
        output.append('pop de')

    output.append('call __STORE32')
    REQUIRES.add('store32.asm')

    return output


def _pastoref(ins):
    """ Stores a floating point value into a memory address.
    """
    output = _paddr(ins.quad[1])

    value = ins.quad[2]
    if value[0] == '*':
        value = value[1:]
        indirect = True
    else:
        indirect = False

    try:
        if indirect:
            value = int(value) & 0xFFFF  # Immediate?
            output.append('push hl')
            output.append('ld hl, %i' % value)
            output.append('call __ILOADF')
            output.append('ld a, c')
            output.append('ld b, h')
            output.append('ld c, l')  # BC = Lower 16 bits, A = Exp
            output.append('pop hl')     # Recovers pointer
            REQUIRES.add('iloadf.asm')
        else:
            value = float(value)  # Immediate?
            C, DE, HL = fp.immediate_float(value)
            output.append('ld a, %s' % C)
            output.append('ld de, %s' % DE)
            output.append('ld bc, %s' % HL)
    except ValueError:
        output.append('pop bc')
        output.append('pop de')
        output.append('ex (sp), hl')  # Preserve HL for STOREF
        output.append('ld a, l')
        output.append('pop hl')

    output.append('call __STOREF')
    REQUIRES.add('storef.asm')

    return output


def _pastorestr(ins):
    """ Stores a string value into a memory address.
    It copies content of 2nd operand (string), into 1st, reallocating
    dynamic memory for the 1st str. These instruction DOES ALLOW
    immediate strings for the 2nd parameter, starting with '#'.
    """
    output = _paddr(ins.quad[1])
    temporal = False
    value = ins.quad[2]

    indirect = value[0] == '*'
    if indirect:
        value = value[1:]

    immediate = value[0]
    if immediate:
        value = value[1:]

    if value[0] == '_':
        if indirect:
            if immediate:
                output.append('ld de, (%s)' % value)
            else:
                output.append('ld de, (%s)' % value)
                output.append('call __LOAD_DE_DE')
                REQUIRES.add('lddede.asm')
        else:
            if immediate:
                output.append('ld de, %s' % value)
            else:
                output.append('ld de, (%s)' % value)
    else:
        output.append('pop de')
        temporal = True

        if indirect:
            output.append('call __LOAD_DE_DE')
            REQUIRES.add('lddede.asm')

    if not temporal:
        output.append('call __STORE_STR')
        REQUIRES.add('storestr.asm')
    else:  # A value already on dynamic memory
        output.append('call __STORE_STR2')
        REQUIRES.add('storestr2.asm')

    return output



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from unittest import TestCase
from six import StringIO

import arch.zx48k  # noqa initializes arch
import symbols
import api.global_ as gl
import api.config as config
from api.symboltable import SymbolTable
from symbols.type_ import Type
from libzxbpp import zxbpp


class TestSymbolARRAYACCESS(TestCase):
    def setUp(self):
        zxbpp.init()
        l1 = 1
        l2 = 2
        l3 = 3
        l4 = 4
        b = symbols.BOUND(l1, l2)
        c = symbols.BOUND(l3, l4)
        self.bounds = symbols.BOUNDLIST.make_node(None, b, c)
        self.arr = symbols.VARARRAY('test', self.bounds, 1, type_=Type.ubyte)
        self.arg = symbols.ARGLIST(symbols.ARGUMENT(symbols.NUMBER(2, 1, type_=Type.uinteger), 1),
                                   symbols.ARGUMENT(symbols.NUMBER(3, 1, type_=Type.uinteger), 1))
        gl.SYMBOL_TABLE = SymbolTable()
        # Clears stderr and prepares for capturing it
        config.OPTIONS.remove_option('stderr')
        config.OPTIONS.add_option('stderr', None, StringIO())
        config.OPTIONS.add_option_if_not_defined('explicit', None, False)

    @property
    def OUTPUT(self):
        return config.OPTIONS.stderr.value.getvalue()

    def test__init__(self):
        aa = symbols.ARRAYACCESS(self.arr, self.arg, 2)
        self.assertIsInstance(aa, symbols.ARRAYACCESS)

    def test__init__fail(self):
        # First argument must be an instance of VARARRAY
        self.assertRaises(AssertionError, symbols.ARRAYACCESS, 'bla', self.arg, 2)

    def test_entry__getter(self):
        aa = symbols.ARRAYACCESS(self.arr, self.arg, 2)
        self.assertIs(aa.entry, self.arr)

    def test_entry__setter(self):
        aa = symbols.ARRAYACCESS(self.arr, self.arg, 2)
        ar2 = symbols.VARARRAY('test2', self.bounds, 1, type_=Type.ubyte)
        aa.entry = ar2
        self.assertIs(aa.entry, ar2)

    def test_entry__setter_fail(self):
        # entry must be an instance of VARARRAY
        aa = symbols.ARRAYACCESS(self.arr, self.arg, 2)
        self.assertRaises(AssertionError, symbols.ARRAYACCESS.entry.fset, aa, 'blah')

    def test_scope(self):
        aa = symbols.ARRAYACCESS(self.arr, self.arg, 2)
        self.assertEqual(aa.scope, self.arr.scope)

    def test_make_node(self):
        gl.SYMBOL_TABLE.declare_array('test', 1, symbols.TYPEREF(self.arr.type_, 1),
                                      bounds=self.bounds)
        aa = symbols.ARRAYACCESS.make_node('test', self.arg, lineno=2)
        self.assertIsInstance(aa, symbols.ARRAYACCESS)

    def test_make_node_fail(self):
        gl.SYMBOL_TABLE.declare_array('test', 1, symbols.TYPEREF(self.arr.type_, 1),
                                      bounds=self.bounds)
        self.arg = symbols.ARGLIST(symbols.ARGUMENT(symbols.NUMBER(2, 1), 1))
        aa = symbols.ARRAYACCESS.make_node('test', self.arg, lineno=2)
        self.assertIsNone(aa)
        self.assertEqual(self.OUTPUT, "(stdin):2: error: Array 'test' has 2 dimensions, not 1\n")

    def test_make_node_warn(self):
        gl.SYMBOL_TABLE.declare_array('test', 1, symbols.TYPEREF(self.arr.type_, 1),
                                      bounds=self.bounds)
        self.arg[1] = symbols.ARGUMENT(symbols.NUMBER(9, 1), 1)
        aa = symbols.ARRAYACCESS.make_node('test', self.arg, lineno=2)
        self.assertIsNotNone(aa)
        self.assertEqual(self.OUTPUT, "(stdin):2: warning: Array 'test' subscript out of range\n")

    def test_offset(self):
        gl.SYMBOL_TABLE.declare_array('test', 1, symbols.TYPEREF(self.arr.type_, 1),
                                      bounds=self.bounds)
        aa = symbols.ARRAYACCESS.make_node('test', self.arg, lineno=2)
        self.assertIsInstance(aa, symbols.ARRAYACCESS)
        self.assertIsNotNone(aa.offset)
        self.assertEqual(aa.offset, 2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4:et:sw=4:

# ----------------------------------------------------------------------
# Copyleft (K), Jose M. Rodriguez-Rosa (a.k.a. Boriel)
#
# This program is Free Software and is released under the terms of
#                    the GNU General License
# ----------------------------------------------------------------------

from .symbol_ import Symbol
from .number import SymbolNUMBER
from .type_ import SymbolTYPE

from api.check import is_number
from api.check import is_string


class SymbolBUILTIN(Symbol):
    """ Defines an BUILTIN function e.g. INKEY$(), RND() or LEN
    """
    def __init__(self, lineno, fname, type_=None, *operands):
        assert isinstance(lineno, int)
        assert type_ is None or isinstance(type_, SymbolTYPE)
        super(SymbolBUILTIN, self).__init__(*operands)
        self.lineno = lineno
        self.fname = fname
        self.type_ = type_

    @property
    def type_(self):
        if self._type is not None:
            return self._type
        return self.operand.type_

    @type_.setter
    def type_(self, value):
        assert value is None or isinstance(value, SymbolTYPE)
        self._type = value

    @property
    def operand(self):
        return self.children[0] if self.children else None

    @operand.setter
    def operand(self, value):
        assert isinstance(value, Symbol)
        self.children[0] = value

    @property
    def operands(self):
        return self.children

    @operands.setter
    def operands(self, value):
        for x in value:
            assert isinstance(x, Symbol)
        self.children = [x for x in value]

    @property
    def size(self):
        """ sizeof(type)
        """
        if self.type_ is None:
            return 0
        return self.type_.size

    @classmethod
    def make_node(cls, lineno, fname, func=None, type_=None, *operands):
        """ Creates a node for a unary operation. E.g. -x or LEN(a$)

        Parameters:
            -func: function used on constant folding when possible
            -type_: the resulting type (by default, the same as the argument).
                For example, for LEN (str$), result type is 'u16'
                and arg type is 'string'
        """
        if func is not None and len(operands) == 1:  # Try constant-folding
            if is_number(operands[0]) or is_string(operands[0]):  # e.g. ABS(-5)
                return SymbolNUMBER(func(operands[0].value), type_=type_, lineno=lineno)

        return cls(lineno, fname, type_, *operands)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# vim:ts=4:et:sw=4:

# --------------------------------------------------------------
# Copyleft (k) 2008, by Jose M. Rodriguez-Rosa
# (a.k.a. Boriel, http://www.boriel.com)
#
# This module contains array load/store
# intermediate-code traductions
# --------------------------------------------------------------


from .__common import REQUIRES, is_int
from .__f16 import _f16_oper
from .__float import _fpush, _float_oper
from .errors import InvalidICError


def _addr(value):
    ''' Common subroutine for emitting array address
    '''
    output = []

    try:
        indirect = False
        if value[0] == '*':
            indirect = True
            value = value[1:]

        value = int(value) & 0xFFFF
        if indirect:
            output.append('ld hl, (%s)' % str(value))
        else:
            output.append('ld hl, %s' % str(value))

    except ValueError:
        if value[0] == '_':
            output.append('ld hl, %s' % str(value))
            if indirect:
                output.append('ld c, (hl)')
                output.append('inc hl')
                output.append('ld h, (hl)')
                output.append('ld l, c')
        else:
            output.append('pop hl')
            if indirect:
                output.append('ld c, (hl)')
                output.append('inc hl')
                output.append('ld h, (hl)')
                output.append('ld l, c')

    output.append('call __ARRAY')
    REQUIRES.add('array.asm')

    return output


def _aaddr(ins):
    ''' Loads the address of an array element
    into the stack.
    '''
    output = _addr(ins.quad[2])
    output.append('push hl')

    return output


def _aload8(ins):
    ''' Loads an 8 bit value from a memory address
    If 2nd arg. start with '*', it is always treated as
    an indirect value.
    '''
    output = _addr(ins.quad[2])
    output.append('ld a, (hl)')
    output.append('push af')

    return output


def _aload16(ins):
    ''' Loads a 16 bit value from a memory address
    If 2nd arg. start with '*', it is always treated as
    an indirect value.
    '''
    output = _addr(ins.quad[2])

    output.append('ld e, (hl)')
    output.append('inc hl')
    output.append('ld d, (hl)')
    output.append('ex de, hl')
    output.append('push hl')

    return output


def _aload32(ins):
    ''' Load a 32 bit value from a memory address
    If 2nd arg. start with '*', it is always treated as
    an indirect value.
    '''
    output = _addr(ins.quad[2])

    output.append('call __ILOAD32')
    output.append('push de')
    output.append('push hl')

    REQUIRES.add('iload32.asm')

    return output


def _aloadf(ins):
    ''' Loads a floating point value from a memory address.
    If 2nd arg. start with '*', it is always treated as
    an indirect value.
    '''
    output = _addr(ins.quad[2])
    output.append('call __LOADF')
    output.extend(_fpush())

    REQUIRES.add('iloadf.asm')

    return output


def _aloadstr(ins):
    ''' Loads a string value from a memory address.
    '''
    output = _addr(ins.quad[2])

    output.append('call __ILOADSTR')
    output.append('push hl')
    REQUIRES.add('loadstr.asm')

    return output


def _astore8(ins):
    ''' Stores 2º operand content into address of 1st operand.
    1st operand is an array element. Dimensions are pushed into the
    stack.
    Use '*' for indirect store on 1st operand (A pointer to an array)
    '''
    output = _addr(ins.quad[1])
    op = ins.quad[2]

    indirect = op[0] == '*'
    if indirect:
        op = op[1:]

    immediate = op[0] == '#'
    if immediate:
        op = op[1:]

    if is_int(op):
        if indirect:
            if immediate:
                op = str(int(op) & 0xFFFF)  # Truncate to 16bit pointer
                output.append('ld a, (%s)' % op)
            else:
                output.append('ld de, (%s)' % op)
                output.append('ld a, (de)')
        else:
            op = str(int(op) & 0xFF)  # Truncate to byte
            output.append('ld (hl), %s' % op)
            return output

    elif op[0] == '_':
        if indirect:
            if immediate:
                output.append('ld a, (%s)' % op)  # Redundant: *#_id == _id
            else:
                output.append('ld de, (%s)' % op)  # *_id
                output.append('ld a, (de)')
        else:
            if immediate:
                output.append('ld a, %s' % op)  # #_id
            else:
                output.append('ld a, (%s)' % op)  # _id
    else:
        output.append('pop af')  # tn

    output.append('ld (hl), a')

    return output


def _astore16(ins):
    ''' Stores 2º operand content into address of 1st operand.
    store16 a, x =>  *(&a) = x
    Use '*' for indirect store on 1st operand.
    '''
    output = _addr(ins.quad[1])
    op = ins.quad[2]

    indirect = op[0] == '*'
    if indirect:
        op = op[1:]

    immediate = op[0] == '#'
    if immediate:
        op = op[1:]

    if is_int(op):
        op = str(int(op) & 0xFFFF)  # Truncate to 16bit pointer

        if indirect:
            if immediate:
                output.append('ld de, (%s)' % op)
            else:
                output.append('ld de, (%s)' % op)
                output.append('call __LOAD_DE_DE')
                REQUIRES.add('lddede.asm')
        else:
            H = int(op) >> 8
            L = int(op) & 0xFF
            output.append('ld (hl), %i' % L)
            output.append('inc hl')
            output.append('ld (hl), %i' % H)
            return output

    elif op[0] == '_':
        if indirect:
            if immediate:
                output.append('ld de, (%s)' % op)  # redundant: *#_id == _id
            else:
                output.append('ld de, (%s)' % op)  # *_id
                output.append('call __LOAD_DE_DE')
                REQUIRES.add('lddede.asm')
        else:
            if immediate:
                output.append('ld de, %s' % op)
            else:
                output.append('ld de, (%s)' % op)
    else:
        output.append('pop de')

    output.append('ld (hl), e')
    output.append('inc hl')
    output.append('ld (hl), d')

    return output


def _astore32(ins):
    ''' Stores 2º operand content into address of 1st operand.
    store16 a, x =>  *(&a) = x
    '''
    output = _addr(ins.quad[1])

    value = ins.quad[2]
    if value[0] == '*':
        value = value[1:]
        indirect = True
    else:
        indirect = False

    try:
        value = int(ins.quad[2]) & 0xFFFFFFFF  # Immediate?
        if indirect:
            output.append('push hl')
            output.append('ld hl, %i' % (value & 0xFFFF))
            output.append('call __ILOAD32')
            output.append('ld b, h')
            output.append('ld c, l')  # BC = Lower 16 bits
            output.append('pop hl')
            REQUIRES.add('iload32.asm')
        else:
            output.append('ld de, %i' % (value >> 16))
            output.append('ld bc, %i' % (value & 0xFFFF))
    except ValueError:
        output.append('pop bc')
        output.append('pop de')

    output.append('call __STORE32')
    REQUIRES.add('store32.asm')

    return output


def _astoref16(ins):
    ''' Stores 2º operand content into address of 1st operand.
    storef16 a, x =>  *(&a) = x
    '''
    output = _addr(ins.quad[1])

    value = ins.quad[2]
    if value[0] == '*':
        value = value[1:]
        indirect = True
    else:
        indirect = False

    if indirect:
        output.append('push hl')
        output.extend(_f16_oper(ins.quad[2], useBC=True))
        output.append('pop hl')
        REQUIRES.add('iload32.asm')
    else:
        output.extend(_f16_oper(ins.quad[2], useBC=True))

    output.append('call __STORE32')
    REQUIRES.add('store32.asm')

    return output


def _astoref(ins):
    ''' Stores a floating point value into a memory address.
    '''
    output = _addr(ins.quad[1])

    value = ins.quad[2]
    if value[0] == '*':
        value = value[1:]
        indirect = True
    else:
        indirect = False

    if indirect:
        output.append('push hl')
        output.extend(_float_oper(ins.quad[2]))
        output.append('pop hl')
    else:
        output.extend(_float_oper(ins.quad[2]))

    output.append('call __STOREF')
    REQUIRES.add('storef.asm')

    return output


def _astorestr(ins):
    ''' Stores a string value into a memory address.
    It copies content of 2nd operand (string), into 1st, reallocating
    dynamic memory for the 1st str. These instruction DOES ALLOW
    immediate strings for the 2nd parameter, starting with '#'.
    '''
    output = _addr(ins.quad[1])
    op = ins.quad[2]

    indirect = op[0] == '*'
    if indirect:
        op = op[1:]

    immediate = op[0] == '#'
    if immediate:
        op = op[1:]

    temporal = op[0] != '$'
    if not temporal:
        op = op[1:]

    if is_int(op):
        op = str(int(op) & 0xFFFF)
        if indirect:
            if immediate:  # *#<addr> = ld hl, (number)
                output.append('ld de, (%s)' % op)
            else:
                output.append('ld de, (%s)' % op)
                output.append('call __LOAD_DE_DE')
                REQUIRES.add('lddede.asm')
        else:
            # Integer does not make sense here (unless it's a ptr)
            raise InvalidICError(str(ins))

            output.append('ld de, (%s)' % op)
    elif op[0] == '_':  # an identifier
        temporal = False  # Global var is not a temporary string

        if indirect:
            if immediate:  # *#_id = _id
                output.append('ld de, (%s)' % op)
            else:  # *_id
                output.append('ld de, (%s)' % op)
                output.append('call __LOAD_DE_DE')
                REQUIRES.add('lddede.asm')
        else:
            if immediate:
                output.append('ld de, %s' % op)
            else:
                output.append('ld de, (%s)' % op)
    else:  # tn
        output.append('pop de')

        if indirect:
            output.append('call __LOAD_DE_DE')
            REQUIRES.add('lddede.asm')

    if not temporal:
        output.append('call __STORE_STR')
        REQUIRES.add('storestr.asm')
    else:  # A value already on dynamic memory
        output.append('call __STORE_STR2')
        REQUIRES.add('storestr2.asm')

    return output



# -*- coding: utf-8 -*-
import pathlib
from setuptools import setup

packages = [
    'api',
    'arch',
    'arch.zx48k',
    'arch.zx48k.backend',
    'arch.zx48k.optimizer',
    'arch.zx48k.peephole',
    'ast_',
    'libzxbasm',
    'libzxbc',
    'libzxbpp',
    'libzxbpp.prepro',
    'outfmt',
    'parsetab',
    'ply',
    'symbols'
]

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

package_data = {'': ['*'], 'arch.zx48k.peephole': ['opts/*']}

entry_points = {
    'console_scripts': ['zxb = libzxbc.zxb:main',
                        'zxbasm = libzxbasm.zxbasm:main',
                        'zxbc = libzxbc.zxb:main',
                        'zxbpp = libzxbpp.zxbpp:entry_point']
}

setup_kwargs = {
    'name': 'zxbasic',
    'version': '1.12.0',
    'description': "Boriel's ZX BASIC Compiler",
    'classifiers': [
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU Affero General Public License v3',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.8',
    ],
    'long_description_content_type': "text/markdown",
    'long_description': README,
    'author': 'Jose Rodriguez',
    'author_email': 'boriel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://zxbasic.net',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}

setup(**setup_kwargs)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------
# Copyleft (k) 2008, by Jose M. Rodriguez-Rosa
# (a.k.a. Boriel, http://www.boriel.com)
#
# This module contains 8 bit boolean, arithmetic and
# comparation intermediate-code traductions
# --------------------------------------------------------------

from .__common import REQUIRES, is_float, _f_ops
from .__32bit import _add32, _sub32, _lti32, _gti32, _gei32, _lei32, _ne32, _eq32
from .__32bit import _and32, _xor32, _or32, _not32, _neg32, _abs32
from .__float import _negf


# -----------------------------------------------------
# Fixed Point (16.16) bits operands
# -----------------------------------------------------
def f16(op):
    """ Returns a floating point operand converted to 32 bits unsigned int.
    Negative numbers are returned in 2 complement.

    The result is returned in a tuple (DE, HL) => High16 (Int part), Low16 (Decimal part)
    """
    op = float(op)

    negative = op < 0
    if negative:
        op = -op

    DE = int(op)
    HL = int((op - DE) * 2**16) & 0xFFFF
    DE &= 0xFFFF

    if negative:  # Do C2
        DE ^= 0xFFFF
        HL ^= 0xFFFF

        DEHL = ((DE << 16) | HL) + 1
        HL = DEHL & 0xFFFF
        DE = (DEHL >> 16) & 0xFFFF

    return (DE, HL)


def _f16_oper(op1, op2=None, useBC=False, reversed=False):
    """ Returns pop sequence for 32 bits operands
    1st operand in HLDE, 2nd operand remains in the stack

    Now it does support operands inversion calling __SWAP32.

    However, if 1st operand is integer (immediate) or indirect, the stack
    will be rearranged, so it contains a 32 bit pushed parameter value for the
    subroutine to be called.

    If preserveHL is True, then BC will be used instead of HL for lower part
    for the 1st operand.
    """
    output = []

    if op1 is not None:
        op1 = str(op1)

    if op2 is not None:
        op2 = str(op2)

    op = op2 if op2 is not None else op1

    float1 = False  # whether op1 (2nd operand) is float

    indirect = (op[0] == '*')
    if indirect:
        op = op[1:]

    immediate = (op[0] == '#')
    if immediate:
        op = op[1:]

    hl = 'hl' if not useBC and not indirect else 'bc'

    if is_float(op):
        float1 = True
        op = float(op)

        if indirect:
            op = int(op) & 0xFFFF
            if immediate:
                output.append('ld hl, %i' % op)
            else:
                output.append('ld hl, (%i)' % op)

            output.append('call __ILOAD32')
            REQUIRES.add('iload32.asm')

            if preserveHL:  # noqa TODO: it will fail
                output.append('ld b, h')
                output.append('ld c, l')
        else:
            DE, HL = f16(op)
            output.append('ld de, %i' % DE)
            output.append('ld %s, %i' % (hl, HL))
    else:
        if op[0] == '_':
            if immediate:
                output.append('ld %s, %s' % (hl, op))
            else:
                output.append('ld %s, (%s)' % (hl, op))
        else:
            output.append('pop %s' % hl)

        if indirect:
            output.append('call __ILOAD32')
            REQUIRES.add('iload32.asm')

            if preserveHL:  # noqa TODO: it will fail
                output.append('ld b, h')
                output.append('ld c, l')
        else:
            if op[0] == '_':
                output.append('ld de, (%s + 2)' % op)
            else:
                output.append('pop de')

    if op2 is not None:
        op = op1

        indirect = (op[0] == '*')
        if indirect:
            op = op[1:]

        immediate = (op[0] == '#')
        if immediate:
            op = op[1:]

        if is_float(op):
            op = float(op)

            if indirect:
                op = int(op)
                output.append('exx')
                if immediate:
                    output.append('ld hl, %i' % (op & 0xFFFF))
                else:
                    output.append('ld hl, (%i)' % (op & 0xFFFF))

                output.append('call __ILOAD32')
                output.append('push de')
                output.append('push hl')
                output.append('exx')
                REQUIRES.add('iload32.asm')
            else:
                DE, HL = f16(op)
                output.append('ld bc, %i' % DE)
                output.append('push bc')
                output.append('ld bc, %i' % HL)
                output.append('push bc')
        else:
            if indirect:
                output.append('exx')  # uses alternate set to put it on the stack
                if op[0] == '_':
                    if immediate:
                        output.append('ld hl, %s' % op)
                    else:
                        output.append('ld hl, (%s)' % op)
                else:
                    output.append('pop hl')  # Pointers are only 16 bits ***

                output.append('call __ILOAD32')
                output.append('push de')
                output.append('push hl')
                output.append('exx')
                REQUIRES.add('iload32.asm')
            elif op[0] == '_':  # an address
                if float1 or op1[0] == '_':  # If previous op was constant, we can use hl in advance
                    tmp = output
                    output = []
                    output.append('ld hl, (%s + 2)' % op)
                    output.append('push hl')
                    output.append('ld hl, (%s)' % op)
                    output.append('push hl')
                    output.extend(tmp)
                else:
                    output.append('ld bc, (%s + 2)' % op)
                    output.append('push bc')
                    output.append('ld bc, (%s)' % op)
                    output.append('push bc')
            else:
                pass  # 2nd operand remains in the stack

    if op2 is not None and reversed:
        output.append('call __SWAP32')
        REQUIRES.add('swap32.asm')

    return output


def _f16_to_32bit(ins):
    """ If any of the operands within the ins(truction) are numeric,
    convert them to its 32bit representation, otherwise leave them
    as they are.
    """
    ins.quad = [x for x in ins.quad]
    for i in range(2, len(ins.quad)):
        if is_float(ins.quad[i]):
            de, hl = f16(ins.quad[i])
            ins.quad[i] = str((de << 16) | hl)

    ins.quad = tuple(ins.quad)
    return ins


def _addf16(ins):
    """ Pops last 2 bytes from the stack and adds them.
    Then push the result onto the stack.

    Optimizations:
      * If any of the operands is ZERO,
        then do NOTHING: A + 0 = 0 + A = A
    """
    return _add32(_f16_to_32bit(ins))


def _subf16(ins):
    """ Pops last 2 dwords from the stack and subtract them.
    Then push the result onto the stack.
    NOTE: The operation is TOP[0] = TOP[-1] - TOP[0]

    If TOP[0] is 0, nothing is done
    """
    return _sub32(_f16_to_32bit(ins))


def _mulf16(ins):
    """ Multiplies 2 32bit (16.16) fixed point numbers. The result is pushed onto the stack.
    """
    op1, op2 = tuple(ins.quad[2:])

    if _f_ops(op1, op2) is not None:
        op1, op2 = _f_ops(op1, op2)

        if op2 == 1:  # A * 1 => A
            output = _f16_oper(op1)
            output.append('push de')
            output.append('push hl')
            return output

        if op2 == -1:
            return _neg32(ins)

        output = _f16_oper(op1)
        if op2 == 0:
            output.append('ld hl, 0')
            output.append('ld e, h')
            output.append('ld d, l')
            output.append('push de')
            output.append('push hl')
            return output

    output = _f16_oper(op1, str(op2))
    output.append('call __MULF16')
    output.append('push de')
    output.append('push hl')
    REQUIRES.add('mulf16.asm')
    return output


def _divf16(ins):
    """ Divides 2 32bit (16.16) fixed point numbers. The result is pushed onto the stack.

        Optimizations:

         * If 2nd operand is 1, do nothing
         * If 2nd operand is -1, do NEG32
    """
    op1, op2 = tuple(ins.quad[2:])

    if is_float(op2):
        if float(op2) == 1:
            output = _f16_oper(op1)
            output.append('push de')
            output.append('push hl')
            return output

        if float(op2) == -1:
            return _negf(ins)

    rev = not is_float(op1) and op1[0] != 't' and op2[0] == 't'

    output = _f16_oper(op1, op2, reversed=rev)
    output.append('call __DIVF16')
    output.append('push de')
    output.append('push hl')
    REQUIRES.add('divf16.asm')
    return output


def _modf16(ins):
    """ Reminder of div. 2 32bit (16.16) fixed point numbers. The result is pushed onto the stack.
        Optimizations:
         * If 2nd op is 1. Returns 0
    """
    op1, op2 = tuple(ins.quad[2:])

    if is_float(op2) and float(op2) == 1:
        output = _f16_oper(op1)
        output.append('ld hl, 0')
        output.append('push hl')
        output.append('push hl')
        return output

    rev = not is_float(op1) and op1[0] != 't' and op2[0] == 't'

    output = _f16_oper(op1, op2, reversed=rev)
    output.append('call __MODF16')
    output.append('push de')
    output.append('push hl')
    REQUIRES.add('modf16.asm')
    return output


def _negf16(ins):
    """ Negates (arithmetic) top of the stack (Fixed point in DE.HL)

        Fixed point signed version
    """
    return _neg32(_f16_to_32bit(ins))


def _ltf16(ins):
    """ Compares & pops top 2 operands out of the stack, and checks
        if the 1st operand < 2nd operand (top of the stack).
        Pushes 0 if False, 1 if True.

        Fixed point signed version
    """
    return _lti32(_f16_to_32bit(ins))


def _gtf16(ins):
    """ Compares & pops top 2 operands out of the stack, and checks
        if the 1st operand > 2nd operand (top of the stack).
        Pushes 0 if False, 1 if True.

        Fixed point signed version
    """
    return _gti32(_f16_to_32bit(ins))


def _lef16(ins):
    """ Compares & pops top 2 operands out of the stack, and checks
        if the 1st operand <= 2nd operand (top of the stack).
        Pushes 0 if False, 1 if True.

        Fixed point signed version
    """
    return _lei32(_f16_to_32bit(ins))


def _gef16(ins):
    """ Compares & pops top 2 operands out of the stack, and checks
        if the 1st operand >= 2nd operand (top of the stack).
        Pushes 0 if False, 1 if True.

        Fixed point signed version
    """
    return _gei32(_f16_to_32bit(ins))


def _eqf16(ins):
    """ Compares & pops top 2 operands out of the stack, and checks
        if the 1st operand == 2nd operand (top of the stack).
        Pushes 0 if False, 1 if True.

        Fixed point signed version
    """
    return _eq32(_f16_to_32bit(ins))


def _nef16(ins):
    """ Compares & pops top 2 operands out of the stack, and checks
        if the 1st operand != 2nd operand (top of the stack).
        Pushes 0 if False, 1 if True.

        Fixed point signed version
    """
    return _ne32(_f16_to_32bit(ins))


def _orf16(ins):
    """ Compares & pops top 2 operands out of the stack, and checks
        if the 1st operand OR (Logical) 2nd operand (top of the stack).
        Pushes 0 if False, 1 if True.

        Fixed point signed version
    """
    return _or32(_f16_to_32bit(ins))


def _xorf16(ins):
    """ Compares & pops top 2 operands out of the stack, and checks
        if the 1st operand XOR (Logical) 2nd operand (top of the stack).
        Pushes 0 if False, 1 if True.

        Fixed point signed version
    """
    return _xor32(_f16_to_32bit(ins))


def _andf16(ins):
    """ Compares & pops top 2 operands out of the stack, and checks
        if the 1st operand AND (Logical) 2nd operand (top of the stack).
        Pushes 0 if False, 1 if True.

        Fixed point signed version
    """
    return _and32(_f16_to_32bit(ins))


def _notf16(ins):
    """ Negates top of the stack (Fixed point in DE.HL)

        Fixed point signed version
    """
    return _not32(_f16_to_32bit(ins))


def _absf16(ins):
    """ Absolute value of top of the stack (Fixed point in DE.HL)

        Fixed point signed version
    """
    return _abs32(_f16_to_32bit(ins))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4:et:sw=4:

# ----------------------------------------------------------------------
# Copyleft (K), Jose M. Rodriguez-Rosa (a.k.a. Boriel)
#
# This program is Free Software and is released under the terms of
#                    the GNU General License
# ----------------------------------------------------------------------

from .symbol_ import Symbol
from .number import SymbolNUMBER
from .string_ import SymbolSTRING
from .typecast import SymbolTYPECAST
from .type_ import SymbolTYPE
from .type_ import Type as TYPE

from api.check import is_number
from api.check import is_string


class SymbolUNARY(Symbol):
    """ Defines an UNARY EXPRESSION e.g. (a + b)
        Only the operator (e.g. 'PLUS') is stored.
    """
    def __init__(self, oper, operand, lineno, type_=None):
        super(SymbolUNARY, self).__init__(operand)
        self.lineno = lineno
        self.operator = oper
        self._type = type_

    @property
    def type_(self):
        if self._type is not None:
            return self._type
        return self.operand.type_

    @property
    def size(self):
        """ sizeof(type)
        """
        if self.type_ is None:
            return 0
        return self.type_.size

    @property
    def operand(self):
        return self.children[0]

    @operand.setter
    def operand(self, value):
        self.children[0] = value

    def __str__(self):
        return '%s(%s)' % (self.operator, self.operand)

    def __repr__(self):
        return '(%s: %s)' % (self.operator, self.operand)

    @classmethod
    def make_node(cls, lineno, operator, operand, func=None, type_=None):
        """ Creates a node for a unary operation. E.g. -x or LEN(a$)

        Parameters:
            -func: lambda function used on constant folding when possible
            -type_: the resulting type (by default, the same as the argument).
                For example, for LEN (str$), result type is 'u16'
                and arg type is 'string'
        """
        assert type_ is None or isinstance(type_, SymbolTYPE)

        if func is not None:  # Try constant-folding
            if is_number(operand):  # e.g. ABS(-5)
                return SymbolNUMBER(func(operand.value), lineno=lineno)
            elif is_string(operand):  # e.g. LEN("a")
                return SymbolSTRING(func(operand.text), lineno=lineno)

        if type_ is None:
            type_ = operand.type_

        if operator == 'MINUS':
            if not type_.is_signed:
                type_ = type_.to_signed()
                operand = SymbolTYPECAST.make_node(type_, operand, lineno)
        elif operator == 'NOT':
            type_ = TYPE.ubyte

        return cls(operator, operand, lineno, type_)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

from api.constants import TYPE
from symbols import NUMBER
from symbols import BASICTYPE
from symbols import CONST


class TestSymbolNUMBER(TestCase):
    def test__init__(self):
        self.assertRaises(AssertionError, NUMBER, 0, lineno=None)
        self.assertRaises(AssertionError, NUMBER, 0, lineno=1, type_='')

        n = NUMBER(0, lineno=1)
        self.assertEqual(n.type_, BASICTYPE(TYPE.ubyte))

        n = NUMBER(-1, lineno=1)
        self.assertEqual(n.type_, BASICTYPE(TYPE.byte_))

        n = NUMBER(256, lineno=1)
        self.assertEqual(n.type_, BASICTYPE(TYPE.uinteger))

        n = NUMBER(-256, lineno=1)
        self.assertEqual(n.type_, BASICTYPE(TYPE.integer))

    def test__cmp__(self):
        n = NUMBER(0, lineno=1)
        m = NUMBER(1, lineno=2)

        self.assertNotEqual(n, m)
        self.assertEqual(n, n)

        self.assertNotEqual(n, 2)
        self.assertEqual(n, 0)
        self.assertGreater(n, -1)
        self.assertLess(n, 1)

        self.assertGreater(m, n)
        self.assertLess(n, m)

    def test__cmp__const(self):
        n = NUMBER(0, lineno=1)
        m = CONST(NUMBER(1, lineno=2), lineno=2)
        m2 = CONST(NUMBER(0, lineno=3), lineno=3)

        self.assertNotEqual(n, m)
        self.assertEqual(n, n)

        self.assertNotEqual(n, 2)
        self.assertEqual(n, 0)
        self.assertGreater(n, -1)
        self.assertLess(n, 1)

        self.assertGreater(m, n)
        self.assertLess(n, m)
        self.assertEqual(n, m2)
        self.assertEqual(m2, n)

    def test__t(self):
        n = NUMBER(3.14, 1)
        self.assertEqual(n.t, '3.14')

    def test__add__num_num(self):
        a = NUMBER(1, 0)
        b = NUMBER(2, 0)
        self.assertEqual((a + b).t, '3')

    def test__add__num_const(self):
        a = NUMBER(1, 0)
        b = CONST(NUMBER(2, 0), 0)
        self.assertEqual((a + b).t, '3')

    def test__add__num_value(self):
        a = NUMBER(1, 0)
        self.assertEqual((a + 2).t, '3')

    def test__radd__num_const(self):
        a = NUMBER(1, 0)
        b = CONST(NUMBER(2, 0), 0)
        self.assertEqual((b + a).t, '3')

    def test__radd__num_value(self):
        a = NUMBER(1, 0)
        self.assertEqual((2 + a).t, '3')

    def test__sub__num_num(self):
        a = NUMBER(1, 0)
        b = NUMBER(2, 0)
        self.assertEqual((a - b).t, '-1')

    def test__sub__num_const(self):
        a = NUMBER(1, 0)
        b = CONST(NUMBER(2, 0), 0)
        self.assertEqual((a - b).t, '-1')

    def test__sub__num_value(self):
        a = NUMBER(1, 0)
        self.assertEqual((a - 2).t, '-1')

    def test__rsub__num_const(self):
        a = NUMBER(2, 0)
        b = CONST(NUMBER(1, 0), 0)
        self.assertEqual((b - a).t, '-1')

    def test__rsub__num_value(self):
        a = NUMBER(2, 0)
        self.assertEqual((1 - a).t, '-1')

    def test__mul__num_num(self):
        a = NUMBER(3, 0)
        b = NUMBER(2, 0)
        self.assertEqual((a * b).t, '6')

    def test__mul__num_const(self):
        a = NUMBER(3, 0)
        b = CONST(NUMBER(2, 0), 0)
        self.assertEqual((a * b).t, '6')

    def test__mul__num_value(self):
        a = NUMBER(3, 0)
        self.assertEqual((a * 2).t, '6')

    def test__rmul__num_const(self):
        a = NUMBER(3, 0)
        b = CONST(NUMBER(2, 0), 0)
        self.assertEqual((b * a).t, '6')

    def test__rmul__num_value(self):
        a = NUMBER(3, 0)
        self.assertEqual((2 * a).t, '6')

    def test__div__num_num(self):
        a = NUMBER(3, 0)
        b = NUMBER(-2, 0)
        self.assertEqual((a / b).t, str(a.value / b.value))

    def test__div__num_const(self):
        a = NUMBER(3, 0)
        b = CONST(NUMBER(-2, 0), 0)
        self.assertEqual((a / b).t, str(a.value / b.expr.value))

    def test__div__num_value(self):
        a = NUMBER(3, 0)
        self.assertEqual((a / -2.0).t, '-1.5')

    def test__rdiv__num_const(self):
        a = CONST(NUMBER(-3, 0), 0)
        b = NUMBER(2, 0)
        self.assertEqual((a / b).t, str(a.expr.value / b.value))

    def test__rdiv__num_value(self):
        a = NUMBER(-2, 0)
        self.assertEqual((3.0 / a).t, '-1.5')

    def test__bor__val_num(self):
        b = NUMBER(5, 0)
        self.assertEqual((3 | b).t, '7')

    def test__bor__num_val(self):
        b = NUMBER(5, 0)
        self.assertEqual((b | 3).t, '7')

    def test__band__num_val(self):
        b = NUMBER(5, 0)
        self.assertEqual((b & 3).t, '1')

    def test__band__val_num(self):
        b = NUMBER(5, 0)
        self.assertEqual((3 & b).t, '1')

    def test__mod__num_val(self):
        b = NUMBER(5, 0)
        self.assertEqual((b % 3).t, '2')

    def test__mod__val_num(self):
        b = NUMBER(3, 0)
        self.assertEqual((5 % b).t, '2')

    def test__le__val_num(self):
        b = NUMBER(3, 0)
        self.assertLessEqual(2, b)

    def test__le__num_num(self):
        a = NUMBER(2, 0)
        b = NUMBER(3, 0)
        self.assertLessEqual(a, b)

    def test__ge__val_num(self):
        b = NUMBER(1, 0)
        self.assertGreaterEqual(2, b)

    def test__ge__num_num(self):
        a = NUMBER(4, 0)
        b = NUMBER(3, 0)
        self.assertGreaterEqual(a, b)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from typing import Optional

import collections
from api.errors import Error

__all__ = ['NotAnAstError', 'Tree']


class NotAnAstError(Error):
    """ Thrown when the "pointer" is not
    an AST, but another thing.
    """
    def __init__(self, instance):
        self.instance = instance
        self.msg = "Object '%s' is not an Ast instance" % str(instance)

    def __str__(self):
        return self.msg


class Tree:
    """ Simple tree implementation
    """
    parent: Optional['Tree'] = None

    class ChildrenList:
        def __init__(self, node: 'Tree'):
            assert isinstance(node, Tree)
            self.parent = node  # Node having this children
            self._children = []

        def __getitem__(self, key):
            if isinstance(key, int):
                return self._children[key]

            result = Tree.ChildrenList(self.parent)
            for x in self._children[key]:
                result.append(x)
            return result

        def __setitem__(self, key, value):
            assert value is None or isinstance(value, Tree)
            if value is not None:
                value.parent = self.parent
            self._children[key] = value

        def __delitem__(self, key):
            self._children[key].parent = None
            del self._children[key]

        def append(self, value):
            assert isinstance(value, Tree)
            value.parent = self.parent
            self._children.append(value)

        def insert(self, pos, value):
            assert isinstance(value, Tree)
            value.parent = self.parent
            self._children.insert(pos, value)

        def pop(self, pos=-1):
            result = self._children.pop(pos)
            result.parent = None
            return result

        def __len__(self):
            return len(self._children)

        def __add__(self, other):
            if not isinstance(other, Tree.ChildrenList):
                assert isinstance(other, collections.Container)

            result = Tree.ChildrenList(self.parent)
            for x in self:
                result.append(x)
            for x in other:
                result.append(x)
            return result

        def __repr__(self):
            return "%s:%s" % (self.parent.__repr__(), str([x.__repr__() for x in self._children]))

    def __init__(self):
        self._children = Tree.ChildrenList(self)

    @property
    def children(self):
        return self._children

    @children.setter
    def children(self, value):
        assert isinstance(value, collections.Iterable)
        while len(self.children):
            self.children.pop()

        self._children = Tree.ChildrenList(self)
        for x in value:
            self.children.append(x)

    def inorder(self):
        """ Traverses the tree in order
        """
        for i in self.children:
            yield from i.inorder()

        yield self

    def preorder(self):
        """ Traverses the tree in preorder
        """
        yield self

        for i in self.children:
            yield from i.preorder()

    def postorder(self):
        """ Traverses the tree in postorder
        """
        for i in range(len(self.children) - 1, -1, -1):
            yield from self.children[i].postorder()

        yield self

    def appendChild(self, node):
        """ Appends the given node to the current children list
        """
        self.children.append(node)

    def prependChild(self, node):
        """ Inserts the given node at the beginning of the children list
        """
        self.children.insert(0, node)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import symbols
import api.global_ as gl


class TestSymbolSTRSLICE(TestCase):
    def setUp(self):
        STR = "ZXBASIC"
        self.str_ = symbols.STRING(STR, 1)
        self.lower = symbols.NUMBER(1, 1, type_=gl.SYMBOL_TABLE.basic_types[gl.STR_INDEX_TYPE])
        self.upper = symbols.NUMBER(2, 1, type_=gl.SYMBOL_TABLE.basic_types[gl.STR_INDEX_TYPE])

    def test__init__(self):
        symbols.STRSLICE(self.str_, self.lower, self.upper, 1)

    def test_string__getter(self):
        s = symbols.STRSLICE(self.str_, self.lower, self.upper, 1)
        self.assertEqual(s.string, self.str_)

    def test_string__setter(self):
        s = symbols.STRSLICE(self.str_, self.lower, self.upper, 1)
        tmp = symbols.STRING(self.str_.value * 2, 1)
        s.string = tmp
        self.assertEqual(s.string, tmp)

    def test_string__setter_fail(self):
        s = symbols.STRSLICE(self.str_, self.lower, self.upper, 1)
        self.assertRaises(AssertionError, symbols.STRSLICE.string.fset, s, 0)

    def test_lower(self):
        s = symbols.STRSLICE(self.str_, self.lower, self.upper, 1)
        self.assertEqual(s.lower, self.lower)

    def test_lower__setter(self):
        s = symbols.STRSLICE(self.str_, self.lower, self.upper, 1)
        s.lower = symbols.NUMBER(44, 1, type_=gl.SYMBOL_TABLE.basic_types[gl.STR_INDEX_TYPE])
        self.assertEqual(s.lower, 44)

    def test_upper(self):
        s = symbols.STRSLICE(self.str_, self.lower, self.upper, 1)
        self.assertEqual(s.upper, self.upper)

    def test_upper__setter(self):
        s = symbols.STRSLICE(self.str_, self.lower, self.upper, 1)
        s.upper = symbols.NUMBER(44, 1, type_=gl.SYMBOL_TABLE.basic_types[gl.STR_INDEX_TYPE])
        self.assertEqual(s.upper, 44)

    def test_make_node(self):
        s = symbols.STRSLICE.make_node(1, self.str_, self.lower, self.upper)
        self.assertIsInstance(s, symbols.STRING)
        self.assertEqual(s.value, 'XB')

    def test_make_node_wrong(self):
        bad_index = symbols.VAR('a', 0, type_=gl.SYMBOL_TABLE.basic_types[gl.TYPE.string])
        s = symbols.STRSLICE.make_node(1, self.str_, bad_index, bad_index)
        self.assertIsNone(s)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from collections import OrderedDict
from api.errmsg import syntax_error_not_constant
from api.errmsg import syntax_error_cant_convert_to_type
from api.debug import __DEBUG__

from api.errors import InvalidCONSTexpr
from api.config import OPTIONS
from api.constants import TYPE
from api.constants import SCOPE
import api.global_ as gl

from symbols.symbol_ import Symbol
from symbols.type_ import Type

from . import backend
import symbols

from api.errors import InvalidOperatorError

from .translatorinstvisitor import TranslatorInstVisitor


class TranslatorVisitor(TranslatorInstVisitor):
    """ This visitor just adds the emit() method.
    """
    # ------------------------------------------------
    # A list of tokens that belongs to temporary
    # ATTR setting
    # ------------------------------------------------
    ATTR = ('INK', 'PAPER', 'BRIGHT', 'FLASH', 'OVER', 'INVERSE', 'BOLD', 'ITALIC')
    ATTR_TMP = tuple(x + '_TMP' for x in ATTR)

    # Local flags
    HAS_ATTR = False

    # Previous Token
    PREV_TOKEN = None

    # Current Token
    CURR_TOKEN = None

    LOOPS = []  # Defined LOOPS
    STRING_LABELS = OrderedDict()
    JUMP_TABLES = []

    # Type code used in DATA
    DATA_TYPES = {
        'str': 1,
        'i8': 2,
        'u8': 3,
        'i16': 4,
        'u16': 5,
        'i32': 6,
        'u32': 7,
        'f16': 8,
        'f': 9
    }

    @classmethod
    def reset(cls):
        cls.LOOPS = []  # Defined LOOPS
        cls.STRING_LABELS = OrderedDict()
        cls.JUMP_TABLES = []

    def add_string_label(self, str_):
        """ Maps ("folds") the given string, returning an unique label ID.
        This allows several constant labels to be initialized to the same address
        thus saving memory space.
        :param str_: the string to map
        :return: the unique label ID
        """
        if self.STRING_LABELS.get(str_, None) is None:
            self.STRING_LABELS[str_] = backend.tmp_label()

        return self.STRING_LABELS[str_]

    @property
    def O_LEVEL(self):
        return OPTIONS.optimization.value

    @staticmethod
    def TYPE(type_):
        """ Converts a backend type (from api.constants)
        to a SymbolTYPE object (taken from the SYMBOL_TABLE).
        If type_ is already a SymbolTYPE object, nothing
        is done.
        """
        if isinstance(type_, symbols.TYPE):
            return type_

        assert TYPE.is_valid(type_)
        return gl.SYMBOL_TABLE.basic_types[type_]

    @staticmethod
    def dumpMemory(MEMORY):
        """ Returns a sequence of Quads
        """
        for x in MEMORY:
            yield str(x)

    # Generic Visitor methods
    def visit_BLOCK(self, node):
        __DEBUG__('BLOCK', 2)
        for child in node.children:
            yield child

    # Visits any temporal attribute
    def visit_ATTR_TMP(self, node):
        yield node.children[0]
        self.ic_fparam(node.children[0].type_, node.children[0].t)
        self.ic_call(node.token, 0)  # Procedure call. Discard return
        ifile = node.token.lower()
        ifile = ifile[:ifile.index('_')]
        backend.REQUIRES.add(ifile + '.asm')

    # This function must be called before emit_strings
    def emit_data_blocks(self):
        if not gl.DATA_IS_USED:
            return  # nothing to do

        for label_, datas in gl.DATAS:
            self.ic_label(label_)
            for d in datas:
                if isinstance(d, symbols.FUNCDECL):
                    type_ = '%02Xh' % (self.DATA_TYPES[self.TSUFFIX(d.type_)] | 0x80)
                    self.ic_data(TYPE.byte_, [type_])
                    self.ic_data(gl.PTR_TYPE, [d.mangled])
                    continue

                self.ic_data(TYPE.byte_, [self.DATA_TYPES[self.TSUFFIX(d.value.type_)]])
                if d.value.type_ == self.TYPE(TYPE.string):
                    lbl = self.add_string_label(d.value.value)
                    self.ic_data(gl.PTR_TYPE, [lbl])
                elif d.value.type_ == self.TYPE(TYPE.fixed):  # Convert to bytes
                    bytes_ = 0xFFFFFFFF & int(d.value.value * 2 ** 16)
                    self.ic_data(TYPE.uinteger, ['0x%04X' % (bytes_ & 0xFFFF), '0x%04X' % (bytes_ >> 16)])
                else:
                    self.ic_data(d.value.type_, [self.traverse_const(d.value)])

        if not gl.DATAS:  # The above loop was not executed, because there's no data
            self.ic_label('__DATA__0')
        else:
            missing_data_labels = set(gl.DATA_LABELS_REQUIRED).difference([x.label.name for x in gl.DATAS])
            for data_label in missing_data_labels:
                self.ic_label(data_label)  # A label reference by a RESTORE beyond the last DATA line

        self.ic_vard('__DATA__END', ['00'])

    def emit_strings(self):
        for str_, label_ in self.STRING_LABELS.items():
            l = '%04X' % (len(str_) & 0xFFFF)  # TODO: Universalize for any arch
            self.ic_vard(label_, [l] + ['%02X' % ord(x) for x in str_])

    def emit_jump_tables(self):
        for table_ in self.JUMP_TABLES:
            self.ic_vard(table_.label, [str(len(table_.addresses))] + ['##' + x.mangled for x in table_.addresses])

    def _visit(self, node):
        self.norm_attr()
        if isinstance(node, Symbol):
            __DEBUG__('Visiting {}'.format(node.token), 1)
            if node.token in self.ATTR_TMP:
                return self.visit_ATTR_TMP(node)

        return TranslatorInstVisitor._visit(self, node)

    def norm_attr(self):
        """ Normalize attr state
        """
        if not self.HAS_ATTR:
            return

        self.HAS_ATTR = False
        self.ic_call('COPY_ATTR', 0)
        backend.REQUIRES.add('copy_attr.asm')

    @staticmethod
    def traverse_const(node):
        """ Traverses a constant and returns an string
        with the arithmetic expression
        """
        if node.token == 'NUMBER':
            return node.t

        if node.token == 'UNARY':
            mid = node.operator
            if mid == 'MINUS':
                result = ' -' + TranslatorVisitor.traverse_const(node.operand)
            elif mid == 'ADDRESS':
                if node.operand.scope == SCOPE.global_ or node.operand.token in ('LABEL', 'FUNCTION'):
                    result = TranslatorVisitor.traverse_const(node.operand)
                else:
                    syntax_error_not_constant(node.operand.lineno)
                    return
            else:
                raise InvalidOperatorError(mid)
            return result

        if node.token == 'BINARY':
            mid = node.operator
            if mid == 'PLUS':
                mid = '+'
            elif mid == 'MINUS':
                mid = '-'
            elif mid == 'MUL':
                mid = '*'
            elif mid == 'DIV':
                mid = '/'
            elif mid == 'MOD':
                mid = '%'
            elif mid == 'POW':
                mid = '^'
            elif mid == 'SHL':
                mid = '>>'
            elif mid == 'SHR':
                mid = '<<'
            else:
                raise InvalidOperatorError(mid)

            return '(%s) %s (%s)' % (TranslatorVisitor.traverse_const(node.left), mid,
                                     TranslatorVisitor.traverse_const(node.right))

        if node.token == 'TYPECAST':
            if node.type_ in (Type.byte_, Type.ubyte):
                return '(' + TranslatorVisitor.traverse_const(node.operand) + ') & 0xFF'
            if node.type_ in (Type.integer, Type.uinteger):
                return '(' + TranslatorVisitor.traverse_const(node.operand) + ') & 0xFFFF'
            if node.type_ in (Type.long_, Type.ulong):
                return '(' + TranslatorVisitor.traverse_const(node.operand) + ') & 0xFFFFFFFF'
            if node.type_ == Type.fixed:
                return '((' + TranslatorVisitor.traverse_const(node.operand) + ') & 0xFFFF) << 16'
            syntax_error_cant_convert_to_type(node.lineno, str(node.operand), node.type_)
            return

        if node.token == 'VARARRAY':
            return node.data_label

        if node.token in ('VAR', 'LABEL', 'FUNCTION'):
            # TODO: Check what happens with local vars and params
            return node.t

        if node.token == 'CONST':
            return TranslatorVisitor.traverse_const(node.expr)

        if node.token == 'ARRAYACCESS':
            return '({} + {})'.format(node.entry.data_label, node.offset)

        raise InvalidCONSTexpr(node)

    @staticmethod
    def check_attr(node, n):
        """ Check if ATTR has to be normalized
        after this instruction has been translated
        to intermediate code.
        """
        if len(node.children) > n:
            return node.children[n]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .binary import BinaryEmitter
from .codeemitter import CodeEmitter
from .tzx import TZX
from .tap import TAP


__all__ = [
    'BinaryEmitter',
    'CodeEmitter',
    'TZX',
    'TAP',
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4:et:sw=4:

# ----------------------------------------------------------------------
# Copyleft (K), Jose M. Rodriguez-Rosa (a.k.a. Boriel)
#
# This program is Free Software and is released under the terms of
#                    the GNU General License
# ----------------------------------------------------------------------

# ---- AST Symbols ----
from .symbol_ import Symbol as SYMBOL
from .arglist import SymbolARGLIST as ARGLIST
from .argument import SymbolARGUMENT as ARGUMENT
from .arrayaccess import SymbolARRAYACCESS as ARRAYACCESS
from .arraydecl import SymbolARRAYDECL as ARRAYDECL
from .arrayload import SymbolARRAYLOAD as ARRAYLOAD
from .asm import SymbolASM as ASM
from .binary import SymbolBINARY as BINARY
from .block import SymbolBLOCK as BLOCK
from .bound import SymbolBOUND as BOUND
from .boundlist import SymbolBOUNDLIST as BOUNDLIST
from .builtin import SymbolBUILTIN as BUILTIN
from .call import SymbolCALL as CALL
from .const import SymbolCONST as CONST
from .funccall import SymbolFUNCCALL as FUNCCALL
from .funcdecl import SymbolFUNCDECL as FUNCDECL
from .function import SymbolFUNCTION as FUNCTION
from .nop import SymbolNOP as NOP
from .number import SymbolNUMBER as NUMBER
from .paramdecl import SymbolPARAMDECL as PARAMDECL
from .paramlist import SymbolPARAMLIST as PARAMLIST
from .sentence import SymbolSENTENCE as SENTENCE
from .string_ import SymbolSTRING as STRING
from .strslice import SymbolSTRSLICE as STRSLICE
from .type_ import SymbolBASICTYPE as BASICTYPE
from .type_ import SymbolTYPE as TYPE
from .type_ import SymbolTYPEREF as TYPEREF
from .typecast import SymbolTYPECAST as TYPECAST
from .unary import SymbolUNARY as UNARY
from .var import SymbolVAR as VAR
from .vararray import SymbolVARARRAY as VARARRAY
from .vardecl import SymbolVARDECL as VARDECL
from .label import SymbolLABEL as LABEL


__all__ = [
    'ARGLIST',
    'ARGUMENT',
    'ARRAYACCESS',
    'ARRAYDECL',
    'ARRAYLOAD',
    'ASM',
    'BASICTYPE',
    'BINARY',
    'BLOCK',
    'BOUND',
    'BOUNDLIST',
    'BUILTIN',
    'CALL',
    'CONST',
    'FUNCCALL',
    'FUNCDECL',
    'FUNCTION',
    'LABEL',
    'NOP',
    'NUMBER',
    'PARAMDECL',
    'PARAMLIST',
    'SENTENCE',
    'STRING',
    'STRSLICE',
    'SYMBOL',
    'TYPE',
    'TYPEREF',
    'TYPECAST',
    'UNARY',
    'VAR',
    'VARARRAY',
    'VARDECL',
]



# -*- coding: utf-8 -*-

import pytest
from libzxbc import zxb
import os

PATH = os.path.realpath(os.path.dirname(os.path.abspath(__file__)))


class EnsureRemoveFile(object):
    """ Ensures a filename is removed if exists after
    a block of code is executed
    """
    def __init__(self, output_file_name):
        self.fname = output_file_name

    def remove_file(self):
        if os.path.isfile(self.fname):
            os.unlink(self.fname)

    def __enter__(self):
        self.remove_file()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.remove_file()


@pytest.fixture
def file_bas():
    return os.path.join(PATH, 'empty.bas')


@pytest.fixture
def file_bin():
    return os.path.join(PATH, 'empty.bin')


def test_compile_only(file_bas, file_bin):
    """ Should not generate a file
    """
    with EnsureRemoveFile(file_bin):
        zxb.main(['--parse-only', file_bas, '-o', file_bin])
        assert not os.path.isfile(file_bin), 'Should not create file "empty.bin"'


def test_org_allows_0xnnnn_format(file_bas, file_bin):
    """ Should allow hexadecimal format 0x in org
    """
    with EnsureRemoveFile(file_bin):
        zxb.main(['--parse-only', '--org', '0xC000', file_bas, '-o', file_bin])
        assert zxb.OPTIONS.org.value == 0xC000, 'Should set ORG to 0xC000'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:et:

# Simple debugging module

import os
import inspect

from .config import OPTIONS

__all__ = ['__DEBUG__', '__LINE__', '__FILE__']

# --------------------- END OF GLOBAL FLAGS ---------------------


def __DEBUG__(msg, level=1):
    if level > OPTIONS.Debug.value:
        return

    line = inspect.getouterframes(inspect.currentframe())[1][2]
    fname = os.path.basename(inspect.getouterframes(inspect.currentframe())[1][1])
    OPTIONS.stderr.value.write('debug: %s:%i %s\n' % (fname, line, msg))


def __LINE__():
    """ Returns current file interpreter line
    """
    return inspect.getouterframes(inspect.currentframe())[1][2]


def __FILE__():
    """ Returns current file interpreter line
    """
    return inspect.currentframe().f_code.co_filename



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os

# Initialization
path = os.path.realpath(os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir))
sys.path.insert(0, path)
